#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
The boiler 2D heat exchanger model consist of a cross flow shell and tube
heat exchanger. 1-D Cross Flow Heat Exchanger Model with wall temperatures,
discretization based on tube rows


The model includes shell and tube rigorous heat transfer calculations and
pressure drop calculations for shell side. Note that this model assumes no
phase transitions (if user requires phase transitions, they need a general
model)
Created on Nov 25th, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import units as pyunits

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock
from idaes.models_extra.power_generation.unit_models.boiler_heat_exchanger_2D import (
    HeatExchangerCrossFlow2D_Header,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    # Add property packages to flowsheet library
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()
    # Add a 2D cross-flow heat exchanger with header model to the flowsheet
    m.fs.unit = HeatExchangerCrossFlow2D_Header(
        tube_side={"property_package": m.fs.prop_water, "has_pressure_change": True},
        shell_side={"property_package": m.fs.prop_fluegas, "has_pressure_change": True},
        finite_elements=5,
        flow_type="counter_current",
        tube_arrangement="in-line",
        tube_side_water_phase="Vap",
        has_radiation=True,
        radial_elements=5,
        tube_inner_diameter=0.035,
        tube_thickness=0.0035,
        has_header=True,
        header_radial_elements=5,
        header_inner_diameter=0.3,
        header_wall_thickness=0.03,
    )

    # Primary Superheater (NETL baseline report)
    ITM = 0.0254  # inch to meter conversion
    # m.fs.unit.tube_di.fix((2.5-2*0.165)*ITM)
    # m.fs.unit.tube_thickness.fix(0.165*ITM)
    m.fs.unit.pitch_x.fix(3 * ITM)
    # gas path transverse width 54.78 ft / number of columns
    m.fs.unit.pitch_y.fix(54.78 / 108 * 12 * ITM)
    m.fs.unit.tube_length_seg.fix(53.13 * 12 * ITM)
    m.fs.unit.tube_nseg.fix(20 * 2)
    m.fs.unit.tube_ncol.fix(108)
    m.fs.unit.tube_inlet_nrow.fix(4)
    m.fs.unit.delta_elevation.fix(50)
    m.fs.unit.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.unit.tube_r_fouling = 0.003131  # (0.03131 - 0.1779 h-ft^2-F/BTU)

    # material inputs
    m.fs.unit.therm_cond_wall = 43.0  # Carbon steel 1% C in W/m/K
    m.fs.unit.dens_wall = 7850  # kg/m3 or 0.284 lb/in3
    m.fs.unit.cp_wall = 510.8  # J/kg-K (0.05 to 0.25 % C)
    m.fs.unit.Young_modulus = 2.00e5  # 200 GPa (29,000 ksi)
    m.fs.unit.Possion_ratio = 0.29
    m.fs.unit.coefficient_therm_expansion = 1.3e-5

    m.fs.unit.emissivity_wall.fix(0.7)
    m.fs.unit.fcorrection_htc_tube.fix(1.0)
    m.fs.unit.fcorrection_htc_shell.fix(1.0)
    m.fs.unit.fcorrection_dp_tube.fix(1.0)
    m.fs.unit.fcorrection_dp_shell.fix(1.0)
    m.fs.unit.temperature_ambient.fix(350.0)
    m.fs.unit.head_insulation_thickness.fix(0.025)
    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 11
    # Check unit config arguments
    assert len(m.fs.unit.config) == 19
    assert m.fs.unit.config.shell_side.has_pressure_change
    assert m.fs.unit.config.tube_side.has_pressure_change
    assert m.fs.unit.tube.config.property_package is m.fs.prop_water


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    #   Set inputs
    h = pyo.value(iapws95.htpx(773.15 * pyunits.K, 2.5449e7 * pyunits.Pa))
    print(h)
    m.fs.unit.tube_inlet.flow_mol[0].fix(24678.26)  # mol/s
    m.fs.unit.tube_inlet.enth_mol[0].fix(h)  # J/mol
    m.fs.unit.tube_inlet.pressure[0].fix(2.5449e7)  # Pascals

    # FLUE GAS Inlet relative to a primary superheater
    FGrate = 5109.76  # mol/s equivalent of ~1930.08 klb/hr
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.unit.shell_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.unit.shell_inlet.temperature[0].fix(1102.335)
    m.fs.unit.shell_inlet.pressure[0].fix(100145)

    initialization_tester(build_unit, dof=0)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_run_unit(build_unit):
    m = build_unit
    assert degrees_of_freedom(m) == 0
    optarg = {"tol": 1e-6, "linear_solver": "ma27", "max_iter": 40}
    solver.options = optarg
    # solve model
    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # energy balance
    assert pytest.approx(0, abs=1e-3) == pyo.value(
        m.fs.unit.tube_inlet.flow_mol[0] * m.fs.unit.tube_inlet.enth_mol[0]
        + sum(
            m.fs.unit.shell_inlet.flow_mol_comp[0, i]
            for i in m.fs.unit.shell.config.property_package.component_list
        )
        * m.fs.unit.shell.properties[0, 0].enth_mol
        - m.fs.unit.tube_outlet.flow_mol[0] * m.fs.unit.tube_outlet.enth_mol[0]
        - sum(
            m.fs.unit.shell_outlet.flow_mol_comp[0, i]
            for i in m.fs.unit.shell.config.property_package.component_list
        )
        * m.fs.unit.shell.properties[0, 1].enth_mol
    )
    # pressure drop
    assert pytest.approx(100134.3247, abs=1e-3) == pyo.value(
        m.fs.unit.shell.properties[0, 1].pressure
    )
    # mass balance
    assert pytest.approx(0, abs=1e-3) == pyo.value(
        sum(
            m.fs.unit.shell_inlet.flow_mol_comp[0, i]
            for i in m.fs.unit.shell.config.property_package.component_list
        )
        + m.fs.unit.tube_inlet.flow_mol[0]
        - m.fs.unit.tube_outlet.flow_mol[0]
        - sum(
            m.fs.unit.shell_outlet.flow_mol_comp[0, i]
            for i in m.fs.unit.shell.config.property_package.component_list
        )
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from .autoscaling import AutoScaler
from .custom_scaler_base import CustomScalerBase, ConstraintScalingScheme
from .scaler_profiling import ScalingProfiler
from .util import (
    scaling_factors_from_json_file,
    scaling_factors_to_json_file,
    scaling_factors_from_dict,
    scaling_factors_to_dict,
    get_scaling_factor,
    get_scaling_suffix,
    del_scaling_factor,
    set_scaling_factor,
    report_scaling_factors,
)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Test the parts of helmholtz_state that aren't hit by other tests in this."""

import pytest

import pyomo.environ as pyo

import idaes.models.properties.general_helmholtz.helmholtz_state as hstate
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    AmountBasis,
)
from idaes.core import FlowsheetBlock


@pytest.mark.unit
def test_set_not_fixed():
    m = pyo.ConcreteModel()
    m.x = pyo.Var()
    m.y = pyo.Var(initialize=1)
    state = {"x": 5}
    hstate._StateBlock._set_not_fixed(m.x, state, key="x", hold=True)
    assert pyo.value(m.x) == 5
    assert m.x.fixed
    hstate._StateBlock._set_not_fixed(m.y, state, key="y", hold=True)
    assert pyo.value(m.y) == 1


@pytest.mark.unit
def test_misc_methods():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock()

    m.fs.prop_water = HelmholtzParameterBlock(
        pure_component="h2o", amount_basis=AmountBasis.MOLE
    )
    m.fs.prop_water_mass = HelmholtzParameterBlock(
        pure_component="h2o", amount_basis=AmountBasis.MASS
    )

    m.fs.sb = m.fs.prop_water.build_state_block()
    m.fs.sb_mass = m.fs.prop_water_mass.build_state_block()

    assert m.fs.sb.model_check() is None
    assert "Molar Flow" in m.fs.sb.define_display_vars()
    assert "Mass Flow" in m.fs.sb.define_display_vars()
    assert "T" in m.fs.sb.define_display_vars()
    assert "P" in m.fs.sb.define_display_vars()
    assert "Vapor Fraction" in m.fs.sb.define_display_vars()
    assert "Molar Enthalpy" in m.fs.sb.define_display_vars()

    assert m.fs.sb.flow_mol in m.fs.sb.extensive_state_vars()
    assert m.fs.sb.pressure in m.fs.sb.intensive_state_vars()

    assert "Molar Flow" in m.fs.sb_mass.define_display_vars()
    assert "Mass Flow" in m.fs.sb_mass.define_display_vars()
    assert "T" in m.fs.sb_mass.define_display_vars()
    assert "P" in m.fs.sb_mass.define_display_vars()
    assert "Vapor Fraction" in m.fs.sb_mass.define_display_vars()
    assert "Mass Enthalpy" in m.fs.sb_mass.define_display_vars()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for methods from Reid, Prausnitz and Poling

All methods and parameters from:

The Properties of Gases & Liquids, 3rd Edition
Reid, Prausnitz and Polling, 1977, McGraw-Hill

All parameter indices based on conventions used by the source

Authors: Andrew Lee, Alejandro Garciadiego
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.RPP3 import *
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {
        "cp_mol_ig_comp_coeff": {
            "A": 7.701,
            "B": 4.595e-4,
            "C": 2.521e-6,
            "D": -0.859e-9,
        },
        "enth_mol_form_vap_comp_ref": (-57.797e3, pyunits.cal / pyunits.mol),
        "entr_mol_form_vap_comp_ref": (45.13, pyunits.cal / pyunits.mol / pyunits.K),
        "pressure_sat_comp_coeff": {"A": 18.3036, "B": 3816.44, "C": -46.13},
    }
    m.params.config.include_enthalpy_of_formation = True

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)
    m.config.include_enthalpy_of_formation = True

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=273.15, units=pyunits.K)
    m.params.pressure_ref = Var(initialize=1e5, units=pyunits.Pa)

    m.params.temperature_crit = Var(initialize=647.3, units=pyunits.K)
    m.params.pressure_crit = Var(initialize=217.6e5, units=pyunits.Pa)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=298.15, units=pyunits.K)
    m.props[1].pressure = Var(initialize=101325, units=pyunits.Pa)

    return m


@pytest.mark.unit
def test_cp_mol_ig_comp(frame):
    RPP3.cp_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_ig_comp_coeff_A, Var)
    assert value(frame.params.cp_mol_ig_comp_coeff_A) == 7.701
    assert isinstance(frame.params.cp_mol_ig_comp_coeff_B, Var)
    assert value(frame.params.cp_mol_ig_comp_coeff_B) == 4.595e-4
    assert isinstance(frame.params.cp_mol_ig_comp_coeff_C, Var)
    assert value(frame.params.cp_mol_ig_comp_coeff_C) == 2.521e-6
    assert isinstance(frame.params.cp_mol_ig_comp_coeff_D, Var)
    assert value(frame.params.cp_mol_ig_comp_coeff_D) == -0.859e-9

    expr = RPP3.cp_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(33.636, abs=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(34.447, abs=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_enth_mol_ig_comp(frame):
    RPP3.enth_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_vap_comp_ref, Var)
    assert value(frame.params.enth_mol_form_vap_comp_ref) == (
        pytest.approx(-241822.6, abs=1e-1)
    )

    expr = RPP3.enth_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(-240983.962, abs=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(-237517.968, abs=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_enth_mol_ig_comp_no_form(frame):
    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False
    RPP3.enth_mol_ig_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_vap_comp_ref")

    expr = RPP3.enth_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(838.686, abs=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(4304.680, abs=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_entr_mol_ig_comp(frame):
    RPP3.entr_mol_ig_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_vap_comp_ref, Var)
    assert value(frame.params.entr_mol_form_vap_comp_ref) == (
        pytest.approx(188.8, abs=1e-1)
    )

    expr = RPP3.entr_mol_ig_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(191.761, abs=1e-3)

    frame.props[1].temperature.value = 400
    assert value(expr) == pytest.approx(201.756, abs=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_pressure_sat_comp(frame):
    RPP3.pressure_sat_comp.build_parameters(frame.params)

    assert isinstance(frame.params.pressure_sat_comp_coeff_A, Var)
    assert value(frame.params.pressure_sat_comp_coeff_A) == 18.3036
    assert isinstance(frame.params.pressure_sat_comp_coeff_B, Var)
    assert value(frame.params.pressure_sat_comp_coeff_B) == 3816.44
    assert isinstance(frame.params.pressure_sat_comp_coeff_C, Var)
    assert value(frame.params.pressure_sat_comp_coeff_C) == -46.13

    expr = RPP3.pressure_sat_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(3143.1125, abs=1e-3)

    frame.props[1].temperature.value = 373.15
    assert value(expr) == pytest.approx(101317, rel=1e-4)

    assert_units_equivalent(expr, pyunits.Pa)


@pytest.mark.unit
def test_pressure_sat_comp_dT(frame):
    RPP3.pressure_sat_comp.build_parameters(frame.params)

    expr = RPP3.pressure_sat_comp.dT_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )

    delta = 1e-4 * pyunits.K
    val = RPP3.pressure_sat_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    val_p = RPP3.pressure_sat_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature + delta
    )

    dPdT = value((val - val_p) / -delta)

    assert value(expr) == pytest.approx(dPdT, 1e-4)

    frame.props[1].temperature.value = 373.15

    val = RPP3.pressure_sat_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    val_p = RPP3.pressure_sat_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature + delta
    )

    dPdT = value((val - val_p) / -delta)

    assert value(expr) == pytest.approx(dPdT, 1e-4)

    assert_units_equivalent(expr, pyunits.Pa / pyunits.K)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IAPWS95 property package for water and steam
"""

__author__ = "John Eslick"

# Import IDAES
from idaes.core import declare_process_block_class

import idaes.logger as idaeslog

# Some of these are convenience imports
# pylint: disable=W0611
from idaes.models.properties.general_helmholtz import (
    helmholtz_available,
    HelmholtzParameterBlockData,
    HelmholtzParameterBlock,
    HelmholtzStateBlockData,
    HelmholtzThermoExpressions,
    PhaseType,
    StateVars,
    AmountBasis,
)

from idaes.models.properties.general_helmholtz.helmholtz_state import _StateBlock


# Logger
_log = idaeslog.getLogger(__name__)


def iapws95_available():
    """Make sure the compiled IAPWS-95 functions are available. Yes, in Windows
    the .so extension is still used.
    """
    return helmholtz_available()


def htpx(T=None, P=None, x=None):
    """
    Convenience function to calculate steam enthalpy from temperature and
    either pressure or vapor fraction. This function can be used for inlet
    streams and initialization where temperature is known instead of enthalpy.
    User must provide values for two of T, P, or x.

    Args:
        T: Temperature with units (between 200 and 3000 K)
        P: Pressure with units (between 1 and 1e9 Pa), None if saturated vapor
        x: Vapor fraction [mol vapor/mol total] (between 0 and 1), None if
        superheated or subcooled

    Returns:
        Total molar enthalpy [J/mol].
    """
    prop = HelmholtzParameterBlock(pure_component="H2O")
    prop.construct()
    return prop.htpx(T=T, p=P, x=x)


@declare_process_block_class("Iapws95ParameterBlock")
class Iapws95ParameterBlockData(HelmholtzParameterBlockData):
    """IAWPS95 Parameter Block class"""

    CONFIG = HelmholtzParameterBlockData.CONFIG()
    CONFIG.pure_component = "H2O"
    # TODO: Might need to refactor this
    # pylint: disable=protected-access
    CONFIG.get("pure_component")._default = "H2O"


@declare_process_block_class(
    "Iapws95StateBlock",
    block_class=_StateBlock,
)
class Iapws95StateBlockData(HelmholtzStateBlockData):
    """IAWPS95 State Block class."""



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Tests that helmholtz specific NTU condenser model."""


import pytest
import pyomo.environ as pyo
import idaes.core
from idaes.models_extra.power_generation.unit_models.helm import HelmNtuCondenser
from idaes.models.properties import iapws95
from idaes.core.solvers import get_solver
from idaes.models.properties.general_helmholtz import helmholtz_available

solver = get_solver()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_condenser_steady():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmNtuCondenser(
        dynamic=False,
        shell={"has_pressure_change": False, "property_package": m.fs.properties},
        tube={"has_pressure_change": False, "property_package": m.fs.properties},
    )

    m.fs.unit.shell_inlet.flow_mol.fix(100)
    m.fs.unit.shell_inlet.pressure[:] = 9000
    m.fs.unit.shell_inlet.enth_mol.fix(30000)

    m.fs.unit.tube_inlet.flow_mol.fix(1200)
    m.fs.unit.tube_inlet.pressure.fix(101325)
    m.fs.unit.tube_inlet.enth_mol.fix(1800)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(500)

    m.fs.unit.initialize(unfix="pressure")
    solver.solve(m)

    assert pyo.value(m.fs.unit.shell_inlet.pressure[0]) == pytest.approx(
        14036.3360, rel=1e-2
    )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_condenser_dynamic():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(
        dynamic=True, time_set=[0, 3], time_units=pyo.units.s
    )
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmNtuCondenser(
        dynamic=False,
        shell={"has_pressure_change": False, "property_package": m.fs.properties},
        tube={"has_pressure_change": False, "property_package": m.fs.properties},
    )

    pyo.TransformationFactory("dae.finite_difference").apply_to(
        m.fs, nfe=4, wrt=m.fs.time, scheme="BACKWARD"
    )

    m.fs.unit.shell_inlet.flow_mol.fix(100)
    m.fs.unit.shell_inlet.pressure[:] = 9000
    m.fs.unit.shell_inlet.enth_mol.fix(30000)

    m.fs.unit.tube_inlet.flow_mol.fix(1200)
    m.fs.unit.tube_inlet.pressure.fix(101325)
    m.fs.unit.tube_inlet.enth_mol.fix(1800)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(500)

    m.fs.unit.initialize(unfix="pressure")
    solver.solve(m)

    for t in m.fs.unit.shell_inlet.pressure:
        assert pyo.value(m.fs.unit.shell_inlet.pressure[t]) == pytest.approx(
            14036.3360, rel=1e-2
        )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_get_stream_table_contents():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(
        dynamic=True, time_set=[0, 3], time_units=pyo.units.s
    )
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmNtuCondenser(
        dynamic=False,
        shell={"has_pressure_change": False, "property_package": m.fs.properties},
        tube={"has_pressure_change": False, "property_package": m.fs.properties},
    )

    stable = m.fs.unit._get_stream_table_contents()

    expected = {
        "Units": {
            "Mass Flow": getattr(pyo.units.pint_registry, "kg/s"),
            "Molar Flow": getattr(pyo.units.pint_registry, "mol/s"),
            "Molar Enthalpy": getattr(pyo.units.pint_registry, "J/mol"),
            "P": getattr(pyo.units.pint_registry, "Pa"),
            "T": getattr(pyo.units.pint_registry, "K"),
            "Vapor Fraction": getattr(pyo.units.pint_registry, "dimensionless"),
        },
        "Hot Inlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Hot Outlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Cold Inlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "Cold Outlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
    }

    assert stable.to_dict() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.common.unittest as unittest
import pytest
import pyomo.environ as pyo
from pyomo.util.check_units import (
    assert_units_equivalent,
    assert_units_consistent,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.models_extra.gas_distribution.properties.natural_gas import (
    NaturalGasParameterBlock,
)

"""
Test for simple ideal methane property package created with generic
property framework.
"""


@pytest.mark.unit
class TestNaturalGasPropertyPackage(unittest.TestCase):
    def test_build(self):
        m = pyo.ConcreteModel()
        m.properties = NaturalGasParameterBlock()
        m.state = m.properties.build_state_block(defined_state=False)
        state = m.state

        # The following tests are very specific to this property package.
        # They test units and values, which may be different for other
        # property packages that are valid for the pipeline models.

        assert_units_equivalent(state.pressure.get_units(), pyo.units.bar)
        assert_units_equivalent(
            state.flow_mol.get_units(), pyo.units.kmol / pyo.units.hr
        )
        assert_units_equivalent(state.temperature.get_units(), pyo.units.K)
        component_list = m.state.config.parameters.component_list
        j = next(iter(component_list))
        self.assertIs(str(state.mole_frac_comp[j].get_units()), "dimensionless")

        mw = m.state.mw
        assert_units_equivalent(pyo.units.get_units(mw), pyo.units.kg / pyo.units.kmol)
        self.assertEqual(pyo.value(mw), 18.0)

        flow_mass = m.state.flow_mass
        assert_units_equivalent(
            pyo.units.get_units(flow_mass), pyo.units.kg / pyo.units.hr
        )
        m.state.flow_mol.set_value(100.0 * pyo.units.kmol / pyo.units.hr)
        self.assertAlmostEqual(pyo.value(flow_mass), 1800.0)

        kJkmolK = pyo.units.kJ / pyo.units.kmol / pyo.units.K
        cp_mol_comp = m.state.cp_mol_comp[j]
        assert_units_equivalent(pyo.units.get_units(cp_mol_comp), kJkmolK)
        self.assertAlmostEqual(pyo.value(cp_mol_comp), 2.34 * 18.0)

        cp_mol = m.state.cp_mol
        assert_units_equivalent(pyo.units.get_units(cp_mol), kJkmolK)
        self.assertAlmostEqual(pyo.value(cp_mol), 2.34 * 18.0)

        kJkgK = pyo.units.kJ / pyo.units.kg / pyo.units.K
        cp_mass = m.state.cp_mass
        assert_units_equivalent(cp_mass, kJkgK)
        self.assertEqual(pyo.value(cp_mass), 2.34)

        cv_mol_comp = m.state.cv_mol_comp[j]
        assert_units_equivalent(pyo.units.get_units(cv_mol_comp), kJkmolK)
        self.assertAlmostEqual(pyo.value(cv_mol_comp), 1.85 * 18.0)

        cv_mol = m.state.cv_mol
        assert_units_equivalent(pyo.units.get_units(cv_mol_comp), kJkmolK)
        self.assertAlmostEqual(pyo.value(cv_mol), 1.85 * 18.0)

        cv_mass = m.state.cv_mass
        assert_units_equivalent(pyo.units.get_units(cv_mass), kJkgK)
        self.assertAlmostEqual(pyo.value(cv_mass), 1.85)

        heat_capacity_ratio = m.state.heat_capacity_ratio
        assert_units_equivalent(
            pyo.units.get_units(heat_capacity_ratio), pyo.units.dimensionless
        )
        self.assertAlmostEqual(pyo.value(heat_capacity_ratio), 2.34 / 1.85)

        speed_of_sound = m.state.speed_of_sound
        speed_of_sound_eq = m.state.speed_of_sound_eq
        assert_units_equivalent(speed_of_sound, pyo.units.m / pyo.units.s)
        assert_units_consistent(speed_of_sound_eq)
        m.state.temperature.set_value(293.15 * pyo.units.K)
        calculate_variable_from_constraint(speed_of_sound, speed_of_sound_eq)
        # 370.2 = (0.8*2.34/1.85*8314*293.15/18.0)**0.5
        self.assertAlmostEqual(pyo.value(speed_of_sound), 370.1628645)

        dens_mol = m.state.dens_mol
        dens_mol_eq = m.state.dens_mol_eq
        assert_units_equivalent(dens_mol, pyo.units.kmol / pyo.units.m**3)
        assert_units_consistent(dens_mol_eq)
        m.state.temperature.set_value(293.15 * pyo.units.K)
        m.state.pressure.set_value(50 * pyo.units.bar)
        calculate_variable_from_constraint(dens_mol, dens_mol_eq)
        # 2.654 = 50*1e2/0.8/8.314/293.15
        self.assertAlmostEqual(pyo.value(dens_mol), 2.5642238411)

        component_list = m.state.config.parameters.component_list
        j = next(iter(component_list))
        dens_mol_comp = m.state.dens_mol_comp[j]
        assert_units_equivalent(
            pyo.units.get_units(dens_mol_comp), pyo.units.kmol / pyo.units.m**3
        )
        self.assertAlmostEqual(pyo.value(dens_mol_comp), 2.5642238411)

        # flow_mol was set to 100 kmol/hr earlier.
        m.state.mole_frac_comp[j].set_value(1.0)
        self.assertEqual(
            pyo.value(m.state.flow_mol_comp[j]),
            100.0,
        )
        assert_units_equivalent(
            pyo.units.get_units(m.state.flow_mol_comp[j]),
            pyo.units.kmol / pyo.units.hr,
        )

        enth_flow = m.state.get_enthalpy_flow_terms("Vap")
        pred_enth_flow = pyo.value(
            (293.15 - 298.15) * m.state.cp_mol * m.state.flow_mol
        )
        self.assertAlmostEqual(
            pyo.value(enth_flow),
            pred_enth_flow,
        )
        kJhr = pyo.units.kJ / pyo.units.hr
        assert_units_equivalent(
            pyo.units.get_units(enth_flow),
            kJhr,
        )

    def test_nominal_density(self):
        m = pyo.ConcreteModel()
        m.properties = NaturalGasParameterBlock()
        self.assertEqual(m.properties.dens_nominal.value, 0.72)
        assert_units_equivalent(
            m.properties.dens_nominal.get_units(),
            pyo.units.kg / pyo.units.m**3,
        )

    def test_compress_fact(self):
        m = pyo.ConcreteModel()
        m.properties = NaturalGasParameterBlock()
        m.state = m.properties.build_state_block()
        self.assertEqual(m.state.compress_fact.value, 0.80)

    def test_temperature_ref(self):
        m = pyo.ConcreteModel()
        m.properties = NaturalGasParameterBlock()
        self.assertEqual(pyo.value(m.properties.temperature_ref), 298.15)
        assert_units_equivalent(
            m.properties.temperature_ref,
            pyo.units.K,
        )

    def test_pprint_noerror(self):
        m = pyo.ConcreteModel()
        m.properties = NaturalGasParameterBlock()
        m.state = m.properties.build_state_block()
        m.state.dens_mol
        m.state.pprint()


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ideal equation of state methods

Author: Andrew Lee
"""
import pytest
from sys import modules

from pyomo.environ import ConcreteModel, Var, units as pyunits, value
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import (
    declare_process_block_class,
    LiquidPhase,
    VaporPhase,
    SolidPhase,
    Solute,
    Solvent,
    Apparent,
)
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.core.util.exceptions import ConfigurationError, PropertyNotSupportedError
from idaes.core.util.constants import Constants as const


# Dummy method for property method calls
def dummy_call(b, j, T):
    return 42.0


def dummy_call2(b, j, T):
    return 7.0


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


@pytest.fixture()
def m():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(
        components={"a": {}, "b": {}, "c": {}},
        phases={
            "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )

    return m


@pytest.fixture()
def m_sol():
    m = ConcreteModel()

    # Dummy params block with a Solid Phase to check phase typing
    m.params = DummyParameterBlock(
        components={"a": {}, "b": {}, "c": {}},
        phases={
            "Sol": {"type": SolidPhase, "equation_of_state": Ideal},
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.build_state_block([1], defined_state=False)

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )

    return m


@pytest.mark.unit
def test_common(m):
    assert Ideal.common(m.props, "foo") is None


@pytest.mark.unit
def test_compress_fact_phase_Liq(m):
    assert Ideal.compress_fact_phase(m.props[1], "Liq") == 0


@pytest.mark.unit
def test_compress_fact_phase_Vap(m):
    assert Ideal.compress_fact_phase(m.props[1], "Vap") == 1


@pytest.mark.unit
def test_compress_fact_phase_sol(m_sol):
    assert Ideal.compress_fact_phase(m_sol.props[1], "Sol") == 0


@pytest.mark.unit
def test_cp_mol_phase(m):
    m.props[1].cp_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.cp_mol_phase(m.props[1], p)) == str(
            sum(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].cp_mol_phase_comp[p, j]
                for j in m.params.component_list
            )
        )


@pytest.mark.unit
def test_cp_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.cp_mol_liq_comp = dummy_call
        m.params.get_component(j).config.cp_mol_ig_comp = dummy_call

        assert str(Ideal.cp_mol_phase_comp(m.props[1], "Liq", j)) == str(42.0)
        assert str(Ideal.cp_mol_phase_comp(m.props[1], "Vap", j)) == str(42.0)


@pytest.mark.unit
def test_cv_mol_phase(m):
    m.props[1].cv_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.cv_mol_phase(m.props[1], p)) == str(
            sum(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].cv_mol_phase_comp[p, j]
                for j in m.params.component_list
            )
        )


@pytest.mark.unit
def test_cv_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.cp_mol_liq_comp = dummy_call
        m.params.get_component(j).config.cp_mol_ig_comp = dummy_call

        assert str(Ideal.cv_mol_phase_comp(m.props[1], "Liq", j)) == str(42.0)
        assert str(Ideal.cv_mol_phase_comp(m.props[1], "Vap", j)) == str(
            42.0
            - pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg
                * pyunits.m**2
                / pyunits.s**2
                / pyunits.mol
                / pyunits.K,
            )
        )


@pytest.mark.unit
def test_heat_capacity_ratio_phase(m):
    m.props[1].cp_mol_phase = Var(m.params.phase_list)
    m.props[1].cv_mol_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Ideal.heat_capacity_ratio_phase(m.props[1], p)) == (
            str(m.props[1].cp_mol_phase[p] / m.props[1].cv_mol_phase[p])
        )


@pytest.mark.unit
def test_dens_mass_phase(m):
    m.props[1].dens_mol_phase = Var(m.params.phase_list)
    m.props[1].mw_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Ideal.dens_mass_phase(m.props[1], p)) == str(
            m.props[1].dens_mol_phase[p] * m.props[1].mw_phase[p]
        )


@pytest.mark.unit
def test_dens_mol_phase_liq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call

    assert str(Ideal.dens_mol_phase(m.props[1], "Liq")) == str(
        1
        / (
            1 / 42 * m.props[1].mole_frac_phase_comp["Liq", "a"]
            + 1 / 42 * m.props[1].mole_frac_phase_comp["Liq", "b"]
            + 1 / 42 * m.props[1].mole_frac_phase_comp["Liq", "c"]
        )
    )


@pytest.mark.unit
def test_dens_mol_phase_vap(m):
    assert str(Ideal.dens_mol_phase(m.props[1], "Vap")) == (
        "props[1].pressure/("
        + str(Ideal.gas_constant(m.props[1]))
        + "*props[1].temperature)"
    )


@pytest.mark.unit
def test_dens_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.dens_mol_sol_comp = dummy_call
        m_sol.params.get_component(j).config.dens_mol_liq_comp = dummy_call

    assert str(Ideal.dens_mol_phase(m_sol.props[1], "Sol")) == str(
        1
        / sum(
            m_sol.props[1].mole_frac_phase_comp["Sol", j] / 42
            for j in m_sol.params.component_list
        )
    )


@pytest.mark.unit
def test_energy_internal_mol_phase(m):
    m.props[1].energy_internal_mol_phase_comp = Var(
        m.params.phase_list, m.params.component_list
    )

    for p in m.params.phase_list:
        assert str(Ideal.energy_internal_mol_phase(m.props[1], p)) == str(
            sum(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].energy_internal_mol_phase_comp[p, j]
                for j in m.params.component_list
            )
        )


@pytest.mark.unit
def test_energy_internal_mol_phase_comp_no_h_form(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = False
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        assert str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Liq", j)) == str(
            42.0
        )
        assert str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Vap", j)) == str(
            42.0
            - pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg
                * pyunits.m**2
                / pyunits.s**2
                / pyunits.mol
                / pyunits.K,
            )
            * (m.props[1].temperature - m.params.temperature_ref)
        )


@pytest.mark.unit
def test_energy_internal_mol_phase_comp_with_h_form(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = True
        m.params.get_component(j).config.elemental_composition = {
            "C": 1,
            "He": 2,
            "O": 4,
        }
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        # For liquid phase, delta(n) should be 4 (2*He + 2*O2)
        assert (
            str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Liq", j))
            == "42.0 - 4.0*("
            + str(Ideal.gas_constant(m.props[1]))
            + ")*params.temperature_ref"
        )
        # For vapor phase, delta(n) should be 3 (2*He + 2*O2 - 1*component)
        assert str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Vap", j)) == str(
            42.0
            - pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg
                * pyunits.m**2
                / pyunits.s**2
                / pyunits.mol
                / pyunits.K,
            )
            * (m.props[1].temperature - m.params.temperature_ref)
            - 3.0
            * pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg
                * pyunits.m**2
                / pyunits.s**2
                / pyunits.mol
                / pyunits.K,
            )
            * m.params.temperature_ref
        )


@pytest.mark.unit
def test_enth_mol_phase(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call2

    assert str(Ideal.enth_mol_phase(m.props[1], "Vap")) == (
        "props[1].mole_frac_phase_comp[Vap,a]*42.0 + "
        "props[1].mole_frac_phase_comp[Vap,b]*42.0 + "
        "props[1].mole_frac_phase_comp[Vap,c]*42.0"
    )
    assert str(Ideal.enth_mol_phase(m.props[1], "Liq")) == str(
        sum(
            m.props[1].mole_frac_phase_comp["Liq", j] * 42.0
            for j in m.params.component_list
        )
        + (m.props[1].pressure - m.params.pressure_ref)
        / m.props[1].dens_mol_phase["Liq"]
    )


@pytest.mark.unit
def test_enth_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call2

    for j in m.params.component_list:
        assert str(Ideal.enth_mol_phase_comp(m.props[1], "Liq", j)) == str(
            42.0
            + (m.props[1].pressure - m.params.pressure_ref)
            / m.props[1].dens_mol_phase["Liq"]
        )
        assert str(Ideal.enth_mol_phase_comp(m.props[1], "Vap", j)) == str(42.0)


@pytest.mark.unit
def test_enth_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.enth_mol_sol_comp = dummy_call
        m_sol.params.get_component(j).config.dens_mol_liq_comp = dummy_call2
        m_sol.params.get_component(j).config.dens_mol_sol_comp = dummy_call2

    for j in m_sol.params.component_list:
        assert str(Ideal.enth_mol_phase_comp(m_sol.props[1], "Sol", j)) == str(
            42.0
            + (m_sol.props[1].pressure - m_sol.params.pressure_ref)
            / m_sol.props[1].dens_mol_phase["Sol"]
        )


@pytest.mark.unit
def test_entr_mol_phase(m):
    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.entr_mol_phase(m.props[1], p)) == str(
            sum(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].entr_mol_phase_comp[p, j]
                for j in m.params.component_list
            )
        )


@pytest.mark.unit
def test_entr_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.entr_mol_liq_comp = dummy_call
        m.params.get_component(j).config.entr_mol_ig_comp = dummy_call

        assert str(Ideal.entr_mol_phase_comp(m.props[1], "Liq", j)) == str(42.0)
        assert str(Ideal.entr_mol_phase_comp(m.props[1], "Vap", j)) == (
            "42.0 - "
            + str(Ideal.gas_constant(m.props[1]))
            + "*log(props[1].mole_frac_phase_comp"
            "[Vap,{}]*props[1].pressure/params.pressure_ref)".format(j)
        )


@pytest.mark.unit
def test_entr_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.entr_mol_sol_comp = dummy_call

        assert str(Ideal.entr_mol_phase_comp(m_sol.props[1], "Sol", j)) == str(42.0)


@pytest.mark.unit
def test_fug_phase_comp_liq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.pressure_sat_comp = dummy_call

        assert str(Ideal.fug_phase_comp(m.props[1], "Liq", j)) == str(
            m.props[1].mole_frac_phase_comp["Liq", j] * 42.0
        )


@pytest.mark.unit
def test_fug_phase_comp_vap(m):
    for j in m.params.component_list:
        assert str(Ideal.fug_phase_comp(m.props[1], "Vap", j)) == str(
            m.props[1].mole_frac_phase_comp["Vap", j] * m.props[1].pressure
        )


@pytest.mark.unit
def test_fug_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_phase_comp(m_sol.props[1], "Sol", "a")


@pytest.mark.unit
def test_fug_phase_comp_liq_eq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.pressure_sat_comp = dummy_call

        assert str(
            Ideal.fug_phase_comp_eq(m.props[1], "Liq", j, ("Vap", "Liq"))
        ) == str(m.props[1].mole_frac_phase_comp["Liq", j] * 42.0)


@pytest.mark.unit
def test_fug_phase_comp_vap_eq(m):
    for j in m.params.component_list:
        assert str(
            Ideal.fug_phase_comp_eq(m.props[1], "Vap", j, ("Vap", "Liq"))
        ) == str(m.props[1].mole_frac_phase_comp["Vap", j] * m.props[1].pressure)


@pytest.mark.unit
def test_fug_phase_comp_invalid_phase_eq(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_phase_comp_eq(m_sol.props[1], "Sol", "a", ("Vap", "Liq"))


@pytest.mark.unit
def test_fug_coeff_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert Ideal.fug_coeff_phase_comp(m.props[1], p, j) == 1


@pytest.mark.unit
def test_fug_coeff_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_coeff_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.unit
def test_gibbs_mol_phase(m):
    m.props[1].gibbs_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.gibbs_mol_phase(m.props[1], p)) == str(
            sum(
                m.props[1].mole_frac_phase_comp[p, j]
                * m.props[1].gibbs_mol_phase_comp[p, j]
                for j in m.params.component_list
            )
        )


@pytest.mark.unit
def test_gibbs_mol_phase_comp(m):
    m.props[1].enth_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)
    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list, m.params.component_list)

    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Ideal.gibbs_mol_phase_comp(m.props[1], p, j)) == str(
                m.props[1].enth_mol_phase_comp[p, j]
                - m.props[1].entr_mol_phase_comp[p, j] * m.props[1].temperature
            )


@pytest.mark.unit
def test_pressure_osm_phase_no_solvent(m):
    with pytest.raises(
        ConfigurationError,
        match="called for pressure_osm, but no solvents were "
        "defined. Osmotic pressure requires at least one component to be "
        "declared as a solvent.",
    ):
        m.props[1].pressure_osm_phase.display()


@pytest.mark.unit
def test_pressure_osm_phase(m):
    # Create a dummy solvent_set
    m.params.solvent_set = ["a"]
    m.props[1].dens_mol_phase = Var(
        m.params.phase_list, initialize=55e3, units=pyunits.mol / pyunits.m**3
    )

    assert_units_equivalent(m.props[1].pressure_osm_phase["Liq"], pyunits.Pa)
    assert pytest.approx(value(const.gas_constant * 300 * 55e3), rel=1e-6) == value(
        m.props[1].pressure_osm_phase["Liq"]
    )
    subexpr = m.props[1].conc_mol_phase_comp["Liq", "b"]
    subexpr += m.props[1].conc_mol_phase_comp["Liq", "c"]
    assert str(m.props[1].pressure_osm_phase["Liq"].expr) == (
        str(Ideal.gas_constant(m.props[1]) * m.props[1].temperature * subexpr)
    )
    assert len(m.props[1].pressure_osm_phase) == 1


@pytest.mark.unit
def test_pressure_osm_phase_w_apparent_component():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(
        components={
            "a": {"type": Solvent},
            "b": {"type": Solute},
            "c": {"type": Apparent, "dissociation_species": {"foo": 2}},
        },
        phases={
            "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )

    m.props[1].dens_mol_phase = Var(
        m.params.phase_list, initialize=55e3, units=pyunits.mol / pyunits.m**3
    )

    assert_units_equivalent(m.props[1].pressure_osm_phase["Liq"], pyunits.Pa)
    # Factor of 1.5 accounts for the fact the c is doubled.
    assert pytest.approx(
        value(const.gas_constant * 300 * 1.5 * 55e3), rel=1e-6
    ) == value(m.props[1].pressure_osm_phase["Liq"])

    subexpr = m.props[1].conc_mol_phase_comp["Liq", "b"]
    subexpr += 2 * m.props[1].conc_mol_phase_comp["Liq", "c"]
    assert str(m.props[1].pressure_osm_phase["Liq"].expr) == (
        str(Ideal.gas_constant(m.props[1]) * m.props[1].temperature * subexpr)
    )
    assert len(m.props[1].pressure_osm_phase) == 1


@pytest.mark.unit
def test_vol_mol_phase_no_methods(m):
    with pytest.raises(
        ConfigurationError,
        match="does not have a method defined to use when "
        "calculating molar volume and density for component a "
        "in phase liq. Each component must define a method for "
        "either vol_mol_liq_comp or dens_mol_liq_comp.",
    ):
        Ideal.vol_mol_phase(m.props[1], "Liq")


@pytest.mark.unit
def test_vol_mol_phase():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(
        components={
            "a": {"dens_mol_liq_comp": dummy_call},
            "b": {"dens_mol_liq_comp": dummy_call},
            "c": {"vol_mol_liq_comp": dummy_call},
        },
        phases={
            "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
            "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        },
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        state_definition=modules[__name__],
        pressure_ref=100000.0,
        temperature_ref=300,
    )

    m.props = m.params.state_block_class([1], defined_state=False, parameters=m.params)

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=0.5
    )

    for p in m.params.phase_list:
        if p == "Vap":
            assert str(Ideal.vol_mol_phase(m.props[1], p)) == (
                str(Ideal.gas_constant(m.props[1]))
                + "*props[1].temperature/props[1].pressure"
            )
        else:
            assert str(Ideal.vol_mol_phase(m.props[1], p)) == str(
                1 / 42.0 * m.props[1].mole_frac_phase_comp["Liq", "a"]
                + 1 / 42.0 * m.props[1].mole_frac_phase_comp["Liq", "b"]
                + 42.0 * m.props[1].mole_frac_phase_comp["Liq", "c"]
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Alampy SurrogateModelTrainer
"""
import pytest
import numpy as np
import pandas as pd
import re
import io
import os
from math import sin, cos, log, exp
from pathlib import Path
from io import StringIO

from pyomo.environ import Var, Constraint, Expression, value
from pyomo.common.tempfiles import TempfileManager

from idaes.core.surrogate.alamopy import (
    AlamoTrainer,
    AlamoSurrogate,
    Modelers,
    Screener,
    alamo,
)
from idaes.core.surrogate.surrogate_block import SurrogateBlock
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.surrogate.metrics import compute_fit_metrics


dirpath = Path(__file__).parent.resolve()


# String representation of json output for testing
jstring = (
    '{"surrogate": {"z1": " z1 == 3.9999999999925446303450 * x1**2 - '
    "4.0000000000020765611453 * x2**2 - "
    "2.0999999999859380039879 * x1**4 + "
    "4.0000000000043112180492 * x2**4 + "
    "0.33333333332782633107172 * x1**6 + "
    '0.99999999999972988273811 * x1*x2"}, '
    '"input_labels": ["x1", "x2"], '
    '"output_labels": ["z1"], '
    '"input_bounds": {"x1": [0, 5], "x2": [0, 10]}}'
)


class TestAlamoTrainer:
    @pytest.fixture
    def alamo_trainer(self):
        data = {"x1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z1": [10, 20, 30, 40]}
        data = pd.DataFrame(data)

        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        return AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=data,
        )

    @pytest.mark.unit
    def test_rbf_deprecation(self):
        data = {"x1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z1": [10, 20, 30, 40]}
        data = pd.DataFrame(data)

        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        with pytest.raises(
            ConfigurationError,
            match="Support for radial basis functions in the ALAMOpy wrapper have been "
            "deprecated due to limitations in the current implementation. "
            "Pull requests are welcome to address this, otherwise we recommend using "
            "Pysmo which includes support for radial basis functions.",
        ):
            AlamoTrainer(
                input_labels=input_labels,
                output_labels=output_labels,
                input_bounds=bnds,
                training_dataframe=data,
                grbfcns=True,
            )

        with pytest.raises(
            ConfigurationError,
            match="Support for radial basis functions in the ALAMOpy wrapper have been "
            "deprecated due to limitations in the current implementation. "
            "Pull requests are welcome to address this, otherwise we recommend using "
            "Pysmo which includes support for radial basis functions.",
        ):
            AlamoTrainer(
                input_labels=input_labels,
                output_labels=output_labels,
                input_bounds=bnds,
                training_dataframe=data,
                rbfparam=1.0,
            )

    @pytest.mark.unit
    def test_get_files(self, alamo_trainer):
        alamo_trainer.config.filename = "foo.alm"
        alamo_trainer._get_files()
        assert alamo_trainer._almfile == "foo.alm"
        assert alamo_trainer._trcfile == "foo.trc"

    @pytest.mark.unit
    def test_get_files_default(self, alamo_trainer):
        alamo_trainer._get_files()

        assert alamo_trainer._almfile is not None
        assert os.path.exists(alamo_trainer._almfile)
        assert (
            str(alamo_trainer._trcfile).split(".")[0]
            == str(alamo_trainer._almfile).split(".")[0]
        )

        alamo_trainer._remove_temp_files()

        # Check that we cleaned-up after ourselves
        assert not os.path.exists(alamo_trainer._almfile)

    @pytest.mark.unit
    def test_get_files_exists(self, alamo_trainer):
        alamo_trainer.config.filename = os.path.join(dirpath, "alamotrace.trc")

        with pytest.raises(
            FileExistsError,
            match="alamotrace.trc already exists. Either choose a new "
            "file name or set overwrite_files = True",
        ):
            alamo_trainer._get_files()

    @pytest.mark.unit
    def test_spaces_in_input(self):
        data = {"x 1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z1": [10, 20, 30, 40]}
        data = pd.DataFrame(data)

        input_labels = ["x 1", "x2"]
        output_labels = ["z1"]
        bnds = {"x 1": (0, 5), "x2": (0, 10)}

        alamo_trainer = AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=data,
        )

        with pytest.raises(
            ValueError,
            match="ALAMO does not support the presence of spaces in variable names. "
            "Found space in input names: x 1.",
        ):
            alamo_trainer._verify_inputs()

    @pytest.mark.unit
    def test_spaces_in_output(self):
        data = {"x1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z 1": [10, 20, 30, 40]}
        data = pd.DataFrame(data)

        input_labels = ["x1", "x2"]
        output_labels = ["z 1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        alamo_trainer = AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=data,
        )

        with pytest.raises(
            ValueError,
            match="ALAMO does not support the presence of spaces in variable names. "
            "Found space in output names: z 1.",
        ):
            alamo_trainer._verify_inputs()

    @pytest.mark.unit
    def test_writer_default(self, alamo_trainer):
        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
        )

    @pytest.mark.unit
    def test_writer_reordered(self):
        # Test case where training data is out of order and has extra columns
        data = {
            "junk": [100, 200, 300, 400],
            "x1": [1, 2, 3, 4],
            "z1": [10, 20, 30, 40],
            "x2": [5, 6, 7, 8],
        }
        data = pd.DataFrame(data)

        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        alamo_trainer = AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=data,
        )

        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
        )

    @pytest.mark.unit
    def test_writer_custom_basis(self, alamo_trainer):
        alamo_trainer.config.custom_basis_functions = ["sin(x1*x2)", "x1*tanh(x2)"]
        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n\n"
            "NCUSTOMBAS 2\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n\n"
            "BEGIN_CUSTOMBAS\n"
            "sin(x1*x2)\n"
            "x1*tanh(x2)\n"
            "END_CUSTOMBAS\n"
        )

    @pytest.mark.unit
    def test_writer_min_max_equal(self, alamo_trainer):
        alamo_trainer._input_bounds = {"x1": (0, 5), "x2": (10, 10)}
        stream = io.StringIO()

        with pytest.raises(
            ConfigurationError,
            match="ALAMO configuration error: upper and lower "
            "bounds on input x2 are equal.",
        ):
            alamo_trainer._write_alm_to_stream(stream=stream)

    @pytest.mark.unit
    def test_writer_min_max_reversed(self, alamo_trainer):
        alamo_trainer._input_bounds = {"x1": (0, 5), "x2": (15, 10)}
        stream = io.StringIO()

        with pytest.raises(
            ConfigurationError,
            match="ALAMO configuration error: upper bound is "
            "less than lower bound for input x2.",
        ):
            alamo_trainer._write_alm_to_stream(stream=stream)

    @pytest.mark.unit
    def test_writer_trace(self, alamo_trainer):
        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream, trace_fname="foo.bar")

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n"
            "TRACEFNAME foo.bar\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
        )

    @pytest.mark.unit
    def test_writer_validation_data(self):
        training_data = {"x1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z1": [10, 20, 30, 40]}
        training_data = pd.DataFrame(training_data)
        validation_data = {"x1": [2.5], "x2": [6.5], "z1": [25]}
        validation_data = pd.DataFrame(validation_data)
        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        alamo_trainer = AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=training_data,
            validation_dataframe=validation_data,
        )

        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n"
            "NVALDATA 1\n\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
            "\nBEGIN_VALDATA\n"
            "2.5 6.5 25\n"
            "END_VALDATA\n"
        )

    @pytest.mark.unit
    def test_writer_validation_data_reordered(self):
        # Test case where validation data is out of order and has extra columns
        training_data = {"x1": [1, 2, 3, 4], "x2": [5, 6, 7, 8], "z1": [10, 20, 30, 40]}
        training_data = pd.DataFrame(training_data)
        validation_data = {"junk": [100], "z1": [25], "x1": [2.5], "x2": [6.5]}
        validation_data = pd.DataFrame(validation_data)
        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        bnds = {"x1": (0, 5), "x2": (0, 10)}

        alamo_trainer = AlamoTrainer(
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=bnds,
            training_dataframe=training_data,
            validation_dataframe=validation_data,
        )

        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n"
            "NVALDATA 1\n\n"
            "linfcns 1\n"
            "constant 1\n"
            "maxtime 1000.0\n"
            "numlimitbasis 1\n\n"
            "TRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
            "\nBEGIN_VALDATA\n"
            "2.5 6.5 25\n"
            "END_VALDATA\n"
        )

    @pytest.mark.unit
    def test_writer_full_config(self, alamo_trainer):
        alamo_trainer.config.xfactor = [1.2, 1.6]
        alamo_trainer.config.xscaling = True
        alamo_trainer.config.scalez = True

        alamo_trainer.config.monomialpower = [0.5, 2]
        alamo_trainer.config.multi2power = [1, 2, 3]
        alamo_trainer.config.multi3power = [1, 2, 3, 4]
        alamo_trainer.config.ratiopower = [1, 2, 3, 4, 5]

        alamo_trainer.config.constant = True
        alamo_trainer.config.linfcns = False
        alamo_trainer.config.expfcns = True
        alamo_trainer.config.logfcns = False
        alamo_trainer.config.sinfcns = True
        alamo_trainer.config.cosfcns = False
        alamo_trainer.config.grbfcns = True
        alamo_trainer.config.rbfparam = 7

        alamo_trainer.config.modeler = Modelers.AICc
        alamo_trainer.config.builder = True
        alamo_trainer.config.backstepper = False
        alamo_trainer.config.convpen = 42
        alamo_trainer.config.screener = Screener.SIS
        alamo_trainer.config.ncvf = 4
        alamo_trainer.config.sismult = 5

        alamo_trainer.config.maxiter = -50
        alamo_trainer.config.maxtime = 500
        alamo_trainer.config.datalimitterms = True
        alamo_trainer.config.maxterms = [-1, 2]
        alamo_trainer.config.minterms = [2, -1]
        alamo_trainer.config.numlimitbasis = False
        alamo_trainer.config.exclude = [1, 0]
        alamo_trainer.config.ignore = 1
        alamo_trainer.config.xisint = [1, 0]
        alamo_trainer.config.zisint = 0

        alamo_trainer.config.tolrelmetric = 8.1
        alamo_trainer.config.tolabsmetric = 9.2
        alamo_trainer.config.tolmeanerror = 10.3
        alamo_trainer.config.tolsse = 11.4

        alamo_trainer.config.mipoptca = 12.5
        alamo_trainer.config.mipoptcr = 13.6
        alamo_trainer.config.linearerror = False
        alamo_trainer.config.GAMS = "foo"
        alamo_trainer.config.GAMSSOLVER = "bar"
        alamo_trainer.config.solvemip = False
        alamo_trainer.config.print_to_screen = True

        stream = io.StringIO()
        alamo_trainer._write_alm_to_stream(stream=stream)

        assert stream.getvalue() == (
            "# IDAES Alamopy input file\n"
            "NINPUTS 2\n"
            "NOUTPUTS 1\n"
            "XLABELS x1 x2\n"
            "ZLABELS z1\n"
            "XMIN 0 0\n"
            "XMAX 5 10\n"
            "NDATA 4\n\n"
            "MONO 2\n"
            "MULTI2 3\n"
            "MULTI3 4\n"
            "RATIOS 5\n"
            "xfactor 1.2 1.6\n"
            "xscaling 1\n"
            "scalez 1\n"
            "monomialpower 0.5 2\n"
            "multi2power 1.0 2.0 3.0\n"
            "multi3power 1.0 2.0 3.0 4.0\n"
            "ratiopower 1.0 2.0 3.0 4.0 5.0\n"
            "expfcns 1\n"
            "linfcns 0\n"
            "logfcns 0\n"
            "sinfcns 1\n"
            "cosfcns 0\n"
            "constant 1\n"
            "modeler 3\n"
            "builder 1\n"
            "backstepper 0\n"
            "convpen 42.0\n"
            "screener 2\n"
            "ncvf 4\n"
            "sismult 5\n"
            "maxtime 500.0\n"
            "maxiter -50\n"
            "datalimitterms 1\n"
            "maxterms -1 2\n"
            "minterms 2 -1\n"
            "numlimitbasis 0\n"
            "exclude 1 0\n"
            "ignore 1\n"
            "xisint 1 0\n"
            "zisint 0\n"
            "tolrelmetric 8.1\n"
            "tolabsmetric 9.2\n"
            "tolmeanerror 10.3\n"
            "tolsse 11.4\n"
            "mipoptca 12.5\n"
            "mipoptcr 13.6\n"
            "linearerror 0\n"
            "GAMS foo\n"
            "GAMSSOLVER bar\n"
            "solvemip 0\n"
            "print_to_screen 1\n"
            "\nTRACE 1\n\n"
            "BEGIN_DATA\n"
            "1 5 10\n"
            "2 6 20\n"
            "3 7 30\n"
            "4 8 40\n"
            "END_DATA\n"
        )

    @pytest.mark.component
    def test_file_writer(self, alamo_trainer):
        alamo_trainer._get_files()
        alamo_trainer._write_alm_file()

        with open(alamo_trainer._almfile, "r") as f:
            fcont = f.read()
        f.close()
        alamo_trainer._remove_temp_files()

        assert fcont == (
            f"# IDAES Alamopy input file\n"
            f"NINPUTS 2\n"
            f"NOUTPUTS 1\n"
            f"XLABELS x1 x2\n"
            f"ZLABELS z1\n"
            f"XMIN 0 0\n"
            f"XMAX 5 10\n"
            f"NDATA 4\n\n"
            f"linfcns 1\n"
            f"constant 1\n"
            f"maxtime 1000.0\n"
            f"numlimitbasis 1\n\n"
            f"TRACE 1\n"
            f"TRACEFNAME {alamo_trainer._trcfile}\n\n"
            f"BEGIN_DATA\n"
            f"1 5 10\n"
            f"2 6 20\n"
            f"3 7 30\n"
            f"4 8 40\n"
            f"END_DATA\n"
        )

        # Check for clean up
        assert not os.path.exists(alamo_trainer._almfile)

    @pytest.mark.unit
    @pytest.mark.skipif(not alamo.available(), reason="ALAMO not available")
    def test_call_alamo_empty(self, alamo_trainer, caplog):
        alamo_trainer._almfile = "test"

        # To be safe, clear the logger record before test
        caplog.clear()

        alamo_trainer._call_alamo()

        for record in caplog.records:
            assert record.levelname == "WARNING"
        assert (
            "ALAMO executable returned non-zero return code. Check the "
            "ALAMO output for more information."
        ) in caplog.text

        assert alamo_trainer._temp_context is None

    @pytest.mark.component
    @pytest.mark.skipif(not alamo.available(), reason="ALAMO not available")
    def test_call_alamo_w_input(self, alamo_trainer):
        cwd = os.getcwd()

        alamo_trainer._temp_context = TempfileManager.new_context()
        alamo_trainer._almfile = os.path.join(dirpath, "alamo_test.alm")
        alamo_trainer._wrkdir = dirpath

        alamo_trainer._temp_context.add_tempfile(
            os.path.join(dirpath, "GUI_trace.trc"), exists=False
        )
        alamo_trainer._temp_context.add_tempfile(
            os.path.join(dirpath, "alamo_test.lst"), exists=False
        )
        rc, almlog = alamo_trainer._call_alamo()
        assert rc == 0
        assert "Normal termination" in almlog

        # Check for clean up
        alamo_trainer._temp_context.release(remove=True)
        assert not os.path.exists(os.path.join(dirpath, "GUI_trace.trc"))
        assert not os.path.exists(os.path.join(dirpath, "alamo_test.lst"))
        assert cwd == os.getcwd()

    @pytest.mark.unit
    def test_read_trace_single(self, alamo_trainer):
        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace.trc")
        trc = alamo_trainer._read_trace_file(alamo_trainer._trcfile)

        mdict = {
            "z1": (
                " z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            )
        }

        assert trc == {
            "filename": "GUI_camel6.alm",
            "NINPUTS": "2",
            "NOUTPUTS": "1",
            "INITIALPOINTS": "10",
            "OUTPUT": {"z1": "1"},
            "SET": "0",
            "INITIALIZER": "3",
            "SAMPLER": "1",
            "MODELER": "1",
            "BUILDER": "1",
            "GREEDYBUILD": "T",
            "BACKSTEPPER": "0",
            "GREEDYBACK": "T",
            "REGULARIZER": "0",
            "SOLVEMIP": "F",
            "SSEOLR": {"z1": "0.602E-28"},
            "SSE": {"z1": "0.976E-23"},
            "RMSE": {"z1": "0.988E-12"},
            "R2": {"z1": "1.00"},
            "ModelSize": {"z1": "6"},
            "BIC": {"z1": "-539."},
            "RIC": {"z1": "32.5"},
            "Cp": {"z1": "2.00"},
            "AICc": {"z1": "-513."},
            "HQC": {"z1": "-543."},
            "MSE": {"z1": "0.325E-23"},
            "SSEp": {"z1": "0.976E-23"},
            "MADp": {"z1": "0.115E-07"},
            "OLRTime": {"z1": "0.46875000E-01"},
            "numOLRs": {"z1": "16384"},
            "OLRoneCalls": {"z1": "15"},
            "OLRoneFails": {"z1": "0"},
            "OLRgsiCalls": {"z1": "0"},
            "OLRgsiFails": {"z1": "0"},
            "OLRdgelCalls": {"z1": "16369"},
            "OLRdgelFails": {"z1": "0"},
            "OLRclrCalls": {"z1": "0"},
            "OLRclrFails": {"z1": "0"},
            "OLRgmsCalls": {"z1": "0"},
            "OLRgmsFails": {"z1": "0"},
            "CLRTime": {"z1": "0.0000000"},
            "numCLRs": {"z1": "0"},
            "MIPTime": {"z1": "0.0000000"},
            "NumMIPs": {"z1": "0"},
            "LassoTime": {"z1": "0.0000000"},
            "Metric1Lasso": {"z1": "0.17976931+309"},
            "Metric2Lasso": {"z1": "0.17976931+309"},
            "LassoSuccess": {"z1": "F"},
            "LassoRed": {"z1": "0.0000000"},
            "nBasInitAct": {"z1": "15"},
            "nBas": {"z1": "15"},
            "SimTime": {"z1": "0.0000000"},
            "SimData": {"z1": "0"},
            "TotData": {"z1": "10"},
            "NdataConv": {"z1": "0"},
            "OtherTime": {"z1": "0.46875000E-01"},
            "NumIters": {"z1": "1"},
            "IterConv": {"z1": "0"},
            "TimeConv": {"z1": "0.0000000"},
            "Step0Time": {"z1": "0.0000000"},
            "Step1Time": {"z1": "0.93750000E-01"},
            "Step2Time": {"z1": "0.0000000"},
            "TotalTime": {"z1": "0.93750000E-01"},
            "AlamoStatus": {"z1": "0"},
            "AlamoVersion": {"z1": "2021.5.8"},
            "Model": mdict,
        }

    @pytest.mark.unit
    def test_read_trace_multi(self, alamo_trainer):
        alamo_trainer._output_labels = ["z1", "z2"]

        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace2.trc")
        trc = alamo_trainer._read_trace_file(alamo_trainer._trcfile)

        mdict = {
            "z1": (
                " z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            ),
            "z2": (
                " z2 == 0.72267799849202937756409E-001 * x1 + "
                "0.68451684753912708791823E-001 * x2 + "
                "1.0677896915911471165117 * x1**2 - "
                "0.70576464806224348258468 * x2**2 - "
                "0.40286283566554989543640E-001 * x1**3 + "
                "0.67785668021684807038607E-002 * x2**5 - "
                "0.14017881460354553180281 * x1**6 + "
                "0.77207049441576647286212 * x2**6 + "
                "0.42143309951518070910481 * x1*x2 - "
                "0.41818729807213093907503E-001"
            ),
        }

        assert trc == {
            "filename": "GUI_camel6_twooutputs.alm",
            "NINPUTS": "2",
            "NOUTPUTS": "2",
            "INITIALPOINTS": "10",
            "OUTPUT": {"z1": "1", "z2": "2"},
            "SET": "0",
            "INITIALIZER": "3",
            "SAMPLER": "1",
            "MODELER": "1",
            "BUILDER": "1",
            "GREEDYBUILD": "T",
            "BACKSTEPPER": "0",
            "GREEDYBACK": "T",
            "REGULARIZER": "0",
            "SOLVEMIP": "F",
            "SSEOLR": {"z1": "0.602E-28", "z2": "0.280E-30"},
            "SSE": {"z1": "0.976E-23", "z2": "0.280E-30"},
            "RMSE": {"z1": "0.988E-12", "z2": "0.167E-15"},
            "R2": {"z1": "1.00", "z2": "1.00"},
            "ModelSize": {"z1": "6", "z2": "10"},
            "BIC": {"z1": "-539.", "z2": "-704."},
            "RIC": {"z1": "32.5", "z2": "54.2"},
            "Cp": {"z1": "2.00", "z2": "10.0"},
            "AICc": {"z1": "-513.", "z2": "-707."},
            "HQC": {"z1": "-543.", "z2": "-710."},
            "MSE": {"z1": "0.325E-23", "z2": "0.280E-30"},
            "SSEp": {"z1": "0.976E-23", "z2": "0.280E-30"},
            "MADp": {"z1": "0.115E-07", "z2": "0.118E-11"},
            "OLRTime": {"z1": "0.15625000E-01", "z2": "0.93750000E-01"},
            "numOLRs": {"z1": "16384", "z2": "30827"},
            "OLRoneCalls": {"z1": "30", "z2": "30"},
            "OLRoneFails": {"z1": "0", "z2": "0"},
            "OLRgsiCalls": {"z1": "0", "z2": "0"},
            "OLRgsiFails": {"z1": "0", "z2": "0"},
            "OLRdgelCalls": {"z1": "47181", "z2": "47181"},
            "OLRdgelFails": {"z1": "0", "z2": "0"},
            "OLRclrCalls": {"z1": "0", "z2": "0"},
            "OLRclrFails": {"z1": "0", "z2": "0"},
            "OLRgmsCalls": {"z1": "0", "z2": "0"},
            "OLRgmsFails": {"z1": "0", "z2": "0"},
            "CLRTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "numCLRs": {"z1": "0", "z2": "0"},
            "MIPTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "NumMIPs": {"z1": "0", "z2": "0"},
            "LassoTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "Metric1Lasso": {"z1": "0.17976931+309", "z2": "0.17976931+309"},
            "Metric2Lasso": {"z1": "0.17976931+309", "z2": "0.17976931+309"},
            "LassoSuccess": {"z1": "F", "z2": "F"},
            "LassoRed": {"z1": "0.0000000", "z2": "0.0000000"},
            "nBasInitAct": {"z1": "15", "z2": "15"},
            "nBas": {"z1": "15", "z2": "15"},
            "SimTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "SimData": {"z1": "0", "z2": "0"},
            "TotData": {"z1": "10", "z2": "10"},
            "NdataConv": {"z1": "0", "z2": "0"},
            "OtherTime": {"z1": "0.31250000E-01", "z2": "0.31250000E-01"},
            "NumIters": {"z1": "1", "z2": "1"},
            "IterConv": {"z1": "0", "z2": "0"},
            "TimeConv": {"z1": "0.0000000", "z2": "0.0000000"},
            "Step0Time": {"z1": "0.0000000", "z2": "0.0000000"},
            "Step1Time": {"z1": "0.15625000", "z2": "0.15625000"},
            "Step2Time": {"z1": "0.0000000", "z2": "0.0000000"},
            "TotalTime": {"z1": "0.46875000E-01", "z2": "0.10937500"},
            "AlamoStatus": {"z1": "0", "z2": "0"},
            "AlamoVersion": {"z1": "2021.5.8", "z2": "2021.5.8"},
            "Model": mdict,
        }

    @pytest.mark.unit
    def test_read_trace_number_mismatch(self, alamo_trainer):
        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace2.trc")
        with pytest.raises(
            RuntimeError,
            match="Mismatch when reading ALAMO trace file. "
            "Expected OUTPUT = 1, found 2.",
        ):
            alamo_trainer._read_trace_file(alamo_trainer._trcfile)

    @pytest.mark.unit
    def test_read_trace_label_mismatch(self, alamo_trainer):
        alamo_trainer._output_labels = ["z1", "z3"]

        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace2.trc")
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Mismatch when reading ALAMO trace file. "
                "Label of output variable in expression "
                "(z2) does not match expected label (z3)."
            ),
        ):
            alamo_trainer._read_trace_file(alamo_trainer._trcfile)

    @pytest.mark.unit
    def test_read_trace_w_validation(self, alamo_trainer):
        alamo_trainer._output_labels = ["z1", "z2"]

        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace_w_validation.trc")
        trc = alamo_trainer._read_trace_file(
            alamo_trainer._trcfile, has_validation_data=True
        )

        mdict = {
            "z1": (
                " z1 ==  - 0.99999999999999689137553 * x1 - "
                "0.20616513343350446289142E-015 * x2 + "
                "1.0000000000000002220446 * x1**2 - x2**2 + "
                "0.99999999999999811262086"
            ),
            "z2": (
                " z2 == 1.0000000000000022204460 * x2 - "
                "1.0000000000000002220446 * x1**2 + x2**2 - "
                "0.72986161565720566182285E-016 * x1**3 - "
                "1.0000000000000091038288"
            ),
        }

        assert trc == {
            "filename": "alamo_error.alm",
            "NINPUTS": "2",
            "NOUTPUTS": "2",
            "INITIALPOINTS": "7",
            "OUTPUT": {"z1": "1", "z2": "2"},
            "SET": "0",
            "INITIALIZER": "3",
            "SAMPLER": "1",
            "MODELER": "1",
            "BUILDER": "1",
            "GREEDYBUILD": "T",
            "BACKSTEPPER": "0",
            "GREEDYBACK": "T",
            "REGULARIZER": "0",
            "SOLVEMIP": "F",
            "SSEOLR": {"z1": "0.347E-27", "z2": "0.247E-27"},
            "SSE": {"z1": "0.347E-27", "z2": "0.247E-27"},
            "RMSE": {"z1": "0.704E-14", "z2": "0.594E-14"},
            "R2": {"z1": "1.00", "z2": "1.00"},
            "ModelSize": {"z1": "5", "z2": "5"},
            "BIC": {"z1": "-446.", "z2": "-449."},
            "RIC": {"z1": "20.8", "z2": "20.8"},
            "Cp": {"z1": "3.00", "z2": "3.00"},
            "AICc": {"z1": "-386.", "z2": "-389."},
            "HQC": {"z1": "-450.", "z2": "-452."},
            "MSE": {"z1": "0.347E-27", "z2": "0.247E-27"},
            "SSEp": {"z1": "0.347E-27", "z2": "0.247E-27"},
            "MADp": {"z1": "0.888E-13", "z2": "0.888E-13"},
            "OLRTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "numOLRs": {"z1": "247", "z2": "247"},
            "OLRoneCalls": {"z1": "16", "z2": "16"},
            "OLRoneFails": {"z1": "0", "z2": "0"},
            "OLRgsiCalls": {"z1": "0", "z2": "0"},
            "OLRgsiFails": {"z1": "0", "z2": "0"},
            "OLRdgelCalls": {"z1": "478", "z2": "478"},
            "OLRdgelFails": {"z1": "0", "z2": "0"},
            "OLRclrCalls": {"z1": "0", "z2": "0"},
            "OLRclrFails": {"z1": "0", "z2": "0"},
            "OLRgmsCalls": {"z1": "0", "z2": "0"},
            "OLRgmsFails": {"z1": "0", "z2": "0"},
            "CLRTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "numCLRs": {"z1": "0", "z2": "0"},
            "MIPTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "NumMIPs": {"z1": "0", "z2": "0"},
            "LassoTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "Metric1Lasso": {"z1": "0.17976931+309", "z2": "0.17976931+309"},
            "Metric2Lasso": {"z1": "0.17976931+309", "z2": "0.17976931+309"},
            "LassoSuccess": {"z1": "F", "z2": "F"},
            "LassoRed": {"z1": "0.0000000", "z2": "0.0000000"},
            "nBasInitAct": {"z1": "8", "z2": "8"},
            "nBas": {"z1": "8", "z2": "8"},
            "SimTime": {"z1": "0.0000000", "z2": "0.0000000"},
            "SimData": {"z1": "0", "z2": "0"},
            "TotData": {"z1": "7", "z2": "7"},
            "NdataConv": {"z1": "0", "z2": "0"},
            "OtherTime": {"z1": "0.18750000", "z2": "0.18750000"},
            "NumIters": {"z1": "1", "z2": "1"},
            "IterConv": {"z1": "0", "z2": "0"},
            "TimeConv": {"z1": "0.0000000", "z2": "0.0000000"},
            "Step0Time": {"z1": "0.0000000", "z2": "0.0000000"},
            "Step1Time": {"z1": "0.15625000E-01", "z2": "0.15625000E-01"},
            "Step2Time": {"z1": "0.0000000", "z2": "0.0000000"},
            "TotalTime": {"z1": "0.0000000", "z2": "0.15625000E-01"},
            "AlamoStatus": {"z1": "0", "z2": "0"},
            "AlamoVersion": {"z1": "2021.12.28", "z2": "2021.12.28"},
            "Model": mdict,
        }

    @pytest.mark.unit
    def test_read_trace_invalid_file(self, alamo_trainer):
        alamo_trainer._output_labels = ["z1", "z2"]

        alamo_trainer._trcfile = os.path.join(dirpath, "foo.trc")

        with pytest.raises(
            FileNotFoundError,
            match="Error occurred when trying to read the ALAMO trace file - "
            "this probably indicates that a trace file was not created by "
            "the ALAMO executable. Please check the ALAMO output logs.",
        ):
            alamo_trainer._read_trace_file(
                alamo_trainer._trcfile, has_validation_data=True
            )

    @pytest.mark.unit
    def test_populate_results(self, alamo_trainer):
        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace.trc")
        trc = alamo_trainer._read_trace_file(alamo_trainer._trcfile)
        alamo_trainer._populate_results(trc)

        mdict = {
            "z1": (
                " z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            )
        }

        assert alamo_trainer._results == {
            "filename": "GUI_camel6.alm",
            "NINPUTS": "2",
            "NOUTPUTS": "1",
            "INITIALPOINTS": "10",
            "OUTPUT": {"z1": "1"},
            "SET": "0",
            "INITIALIZER": "3",
            "SAMPLER": "1",
            "MODELER": "1",
            "BUILDER": "1",
            "GREEDYBUILD": "T",
            "BACKSTEPPER": "0",
            "GREEDYBACK": "T",
            "REGULARIZER": "0",
            "SOLVEMIP": "F",
            "SSEOLR": {"z1": "0.602E-28"},
            "SSE": {"z1": "0.976E-23"},
            "RMSE": {"z1": "0.988E-12"},
            "R2": {"z1": "1.00"},
            "ModelSize": {"z1": "6"},
            "BIC": {"z1": "-539."},
            "RIC": {"z1": "32.5"},
            "Cp": {"z1": "2.00"},
            "AICc": {"z1": "-513."},
            "HQC": {"z1": "-543."},
            "MSE": {"z1": "0.325E-23"},
            "SSEp": {"z1": "0.976E-23"},
            "MADp": {"z1": "0.115E-07"},
            "OLRTime": {"z1": "0.46875000E-01"},
            "numOLRs": {"z1": "16384"},
            "OLRoneCalls": {"z1": "15"},
            "OLRoneFails": {"z1": "0"},
            "OLRgsiCalls": {"z1": "0"},
            "OLRgsiFails": {"z1": "0"},
            "OLRdgelCalls": {"z1": "16369"},
            "OLRdgelFails": {"z1": "0"},
            "OLRclrCalls": {"z1": "0"},
            "OLRclrFails": {"z1": "0"},
            "OLRgmsCalls": {"z1": "0"},
            "OLRgmsFails": {"z1": "0"},
            "CLRTime": {"z1": "0.0000000"},
            "numCLRs": {"z1": "0"},
            "MIPTime": {"z1": "0.0000000"},
            "NumMIPs": {"z1": "0"},
            "LassoTime": {"z1": "0.0000000"},
            "Metric1Lasso": {"z1": "0.17976931+309"},
            "Metric2Lasso": {"z1": "0.17976931+309"},
            "LassoSuccess": {"z1": "F"},
            "LassoRed": {"z1": "0.0000000"},
            "nBasInitAct": {"z1": "15"},
            "nBas": {"z1": "15"},
            "SimTime": {"z1": "0.0000000"},
            "SimData": {"z1": "0"},
            "TotData": {"z1": "10"},
            "NdataConv": {"z1": "0"},
            "OtherTime": {"z1": "0.46875000E-01"},
            "NumIters": {"z1": "1"},
            "IterConv": {"z1": "0"},
            "TimeConv": {"z1": "0.0000000"},
            "Step0Time": {"z1": "0.0000000"},
            "Step1Time": {"z1": "0.93750000E-01"},
            "Step2Time": {"z1": "0.0000000"},
            "TotalTime": {"z1": "0.93750000E-01"},
            "AlamoStatus": {"z1": "0"},
            "AlamoVersion": {"z1": "2021.5.8"},
            "Model": mdict,
        }

    @pytest.mark.unit
    def test_build_surrogate_object(self, alamo_trainer):
        alamo_trainer._trcfile = os.path.join(dirpath, "alamotrace.trc")
        trc = alamo_trainer._read_trace_file(alamo_trainer._trcfile)
        alamo_trainer._populate_results(trc)
        alamo_object = alamo_trainer._build_surrogate_object()

        assert isinstance(alamo_object, AlamoSurrogate)
        assert alamo_object._surrogate_expressions == {
            "z1": (
                " z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            )
        }
        assert alamo_object._input_labels == ["x1", "x2"]
        assert alamo_object._output_labels == ["z1"]
        assert alamo_object._input_bounds == {"x1": (0, 5), "x2": (0, 10)}


class TestAlamoSurrogate:
    @pytest.fixture
    def alm_surr1(self):
        surrogate_expressions = {
            "z1": (
                " z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            )
        }
        input_labels = ["x1", "x2"]
        output_labels = ["z1"]
        input_bounds = {"x1": (0, 5), "x2": (0, 10)}

        alm_surr1 = AlamoSurrogate(
            surrogate_expressions, input_labels, output_labels, input_bounds
        )

        return alm_surr1

    @pytest.mark.unit
    def test_evaluate_surrogate(self, alm_surr1):
        x = [
            -2,
            -1.8,
            -1.6,
            -1.4,
            -1.2,
            -1.0,
            -0.8,
            -0.6,
            -0.4,
            -0.2,
            0,
            0.2,
            0.4,
            0.6,
            0.8,
            1.0,
            1.2,
            1.4,
            1.6,
            1.8,
            2.0,
        ]

        inputs = np.array([np.tile(x, len(x)), np.repeat(x, len(x))])
        inputs = pd.DataFrame(inputs.transpose(), columns=["x1", "x2"])

        out = alm_surr1.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out["z1"][i], rel=1e-8) == (
                3.9999999999925446303450 * inputs["x1"][i] ** 2
                - 4.0000000000020765611453 * inputs["x2"][i] ** 2
                - 2.0999999999859380039879 * inputs["x1"][i] ** 4
                + 4.0000000000043112180492 * inputs["x2"][i] ** 4
                + 0.33333333332782633107172 * inputs["x1"][i] ** 6
                + 0.99999999999972988273811 * inputs["x1"][i] * inputs["x2"][i]
            )

        # Recall evaluate_surrogate to make sure cached fcn works
        out2 = alm_surr1.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out2["z1"][i], rel=1e-8) == (
                3.9999999999925446303450 * inputs["x1"][i] ** 2
                - 4.0000000000020765611453 * inputs["x2"][i] ** 2
                - 2.0999999999859380039879 * inputs["x1"][i] ** 4
                + 4.0000000000043112180492 * inputs["x2"][i] ** 4
                + 0.33333333332782633107172 * inputs["x1"][i] ** 6
                + 0.99999999999972988273811 * inputs["x1"][i] * inputs["x2"][i]
            )

    @pytest.mark.unit
    def test_populate_block(self, alm_surr1):
        blk = SurrogateBlock(concrete=True)

        blk.build_model(alm_surr1)

        assert isinstance(blk.inputs, Var)
        assert blk.inputs["x1"].bounds == (0, 5)
        assert blk.inputs["x2"].bounds == (0, 10)
        assert isinstance(blk.outputs, Var)
        assert blk.outputs["z1"].bounds == (None, None)
        assert isinstance(blk.alamo_constraint, Constraint)
        assert len(blk.alamo_constraint) == 1
        assert str(blk.alamo_constraint["z1"].body) == (
            "outputs[z1] - ("
            "3.9999999999925446*inputs[x1]**2 - "
            "4.000000000002077*inputs[x2]**2 - "
            "2.099999999985938*inputs[x1]**4 + "
            "4.000000000004311*inputs[x2]**4 + "
            "0.33333333332782633*inputs[x1]**6 + "
            "0.9999999999997299*inputs[x1]*inputs[x2])"
        )

    @pytest.mark.unit
    def test_populate_block_expr(self, alm_surr1):
        blk = SurrogateBlock(concrete=True)

        blk.build_model(alm_surr1, as_expression=True)

        assert isinstance(blk.inputs, Var)
        assert blk.inputs["x1"].bounds == (0, 5)
        assert blk.inputs["x2"].bounds == (0, 10)
        assert isinstance(blk.alamo_expression, Expression)
        assert len(blk.alamo_expression) == 1
        x1 = 3
        x2 = 2
        blk.inputs["x1"] = 3
        blk.inputs["x2"] = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        assert value(blk.alamo_expression["z1"]) == pytest.approx(z1)

    @pytest.fixture
    def alm_surr2(self):
        surrogate_expressions = {
            "z1": (
                "z1 == 3.9999999999925446303450 * x1**2 - "
                "4.0000000000020765611453 * x2**2 - "
                "2.0999999999859380039879 * x1**4 + "
                "4.0000000000043112180492 * x2**4 + "
                "0.33333333332782633107172 * x1**6 + "
                "0.99999999999972988273811 * x1*x2"
            ),
            "z2": (
                "z2 == 0.72267799849202937756409E-001 * x1 + "
                "0.68451684753912708791823E-001 * x2 + "
                "1.0677896915911471165117 * x1**2 - "
                "0.70576464806224348258468 * x2**2 - "
                "0.40286283566554989543640E-001 * x1**3 + "
                "0.67785668021684807038607E-002 * x2**5 - "
                "0.14017881460354553180281 * x1**6 + "
                "0.77207049441576647286212 * x2**6 + "
                "0.42143309951518070910481 * x1*x2 - "
                "0.41818729807213093907503E-001"
            ),
        }
        input_labels = ["x1", "x2"]
        output_labels = ["z1", "z2"]

        alm_surr2 = AlamoSurrogate(surrogate_expressions, input_labels, output_labels)

        return alm_surr2

    @pytest.mark.unit
    def test_evaluate_surrogate_multi(self, alm_surr2):
        x = [
            -2,
            -1.8,
            -1.6,
            -1.4,
            -1.2,
            -1.0,
            -0.8,
            -0.6,
            -0.4,
            -0.2,
            0,
            0.2,
            0.4,
            0.6,
            0.8,
            1.0,
            1.2,
            1.4,
            1.6,
            1.8,
            2.0,
        ]

        inputs = np.array([np.tile(x, len(x)), np.repeat(x, len(x))])
        inputs = pd.DataFrame(inputs.transpose(), columns=["x1", "x2"])

        out = alm_surr2.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out["z1"][i], rel=1e-8) == (
                3.9999999999925446303450 * inputs["x1"][i] ** 2
                - 4.0000000000020765611453 * inputs["x2"][i] ** 2
                - 2.0999999999859380039879 * inputs["x1"][i] ** 4
                + 4.0000000000043112180492 * inputs["x2"][i] ** 4
                + 0.33333333332782633107172 * inputs["x1"][i] ** 6
                + 0.99999999999972988273811 * inputs["x1"][i] * inputs["x2"][i]
            )
            assert pytest.approx(out["z2"][i], rel=1e-8) == (
                0.72267799849202937756409e-001 * inputs["x1"][i]
                + 0.68451684753912708791823e-001 * inputs["x2"][i]
                + 1.0677896915911471165117 * inputs["x1"][i] ** 2
                - 0.70576464806224348258468 * inputs["x2"][i] ** 2
                - 0.40286283566554989543640e-001 * inputs["x1"][i] ** 3
                + 0.67785668021684807038607e-002 * inputs["x2"][i] ** 5
                - 0.14017881460354553180281 * inputs["x1"][i] ** 6
                + 0.77207049441576647286212 * inputs["x2"][i] ** 6
                + 0.42143309951518070910481 * inputs["x1"][i] * inputs["x2"][i]
                - 0.41818729807213093907503e-001
            )

        # Recall evaluate_surrogate to make sure cached fcn works
        out2 = alm_surr2.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out2["z1"][i], rel=1e-8) == (
                3.9999999999925446303450 * inputs["x1"][i] ** 2
                - 4.0000000000020765611453 * inputs["x2"][i] ** 2
                - 2.0999999999859380039879 * inputs["x1"][i] ** 4
                + 4.0000000000043112180492 * inputs["x2"][i] ** 4
                + 0.33333333332782633107172 * inputs["x1"][i] ** 6
                + 0.99999999999972988273811 * inputs["x1"][i] * inputs["x2"][i]
            )
            assert pytest.approx(out2["z2"][i], rel=1e-8) == (
                0.72267799849202937756409e-001 * inputs["x1"][i]
                + 0.68451684753912708791823e-001 * inputs["x2"][i]
                + 1.0677896915911471165117 * inputs["x1"][i] ** 2
                - 0.70576464806224348258468 * inputs["x2"][i] ** 2
                - 0.40286283566554989543640e-001 * inputs["x1"][i] ** 3
                + 0.67785668021684807038607e-002 * inputs["x2"][i] ** 5
                - 0.14017881460354553180281 * inputs["x1"][i] ** 6
                + 0.77207049441576647286212 * inputs["x2"][i] ** 6
                + 0.42143309951518070910481 * inputs["x1"][i] * inputs["x2"][i]
                - 0.41818729807213093907503e-001
            )

    @pytest.mark.unit
    def test_populate_block_multi(self, alm_surr2):
        blk = SurrogateBlock(concrete=True)

        blk.build_model(alm_surr2)

        assert isinstance(blk.inputs, Var)
        assert blk.inputs["x1"].bounds == (None, None)
        assert blk.inputs["x2"].bounds == (None, None)
        assert isinstance(blk.outputs, Var)
        assert blk.outputs["z1"].bounds == (None, None)
        assert blk.outputs["z2"].bounds == (None, None)
        assert isinstance(blk.alamo_constraint, Constraint)
        assert len(blk.alamo_constraint) == 2
        assert str(blk.alamo_constraint["z1"].body) == (
            "outputs[z1] - ("
            "3.9999999999925446*inputs[x1]**2 - "
            "4.000000000002077*inputs[x2]**2 - "
            "2.099999999985938*inputs[x1]**4 + "
            "4.000000000004311*inputs[x2]**4 + "
            "0.33333333332782633*inputs[x1]**6 + "
            "0.9999999999997299*inputs[x1]*inputs[x2])"
        )
        assert str(blk.alamo_constraint["z2"].body) == (
            "outputs[z2] - ("
            "0.07226779984920294*inputs[x1] + "
            "0.06845168475391271*inputs[x2] + "
            "1.0677896915911471*inputs[x1]**2 - "
            "0.7057646480622435*inputs[x2]**2 - "
            "0.04028628356655499*inputs[x1]**3 + "
            "0.006778566802168481*inputs[x2]**5 - "
            "0.14017881460354553*inputs[x1]**6 + "
            "0.7720704944157665*inputs[x2]**6 + "
            "0.4214330995151807*inputs[x1]*inputs[x2] - "
            "0.041818729807213094)"
        )

    @pytest.fixture
    def alm_surr3(self):
        surrogate_expressions = {
            "z1": (" z1 == 2*sin(x1**2) - 3*cos(x2**3) - " "4*ln(x1**4) + 5*exp(x2**5)")
        }
        input_labels = ["x1", "x2"]
        output_labels = ["z1"]

        alm_surr3 = AlamoSurrogate(surrogate_expressions, input_labels, output_labels)

        return alm_surr3

    @pytest.mark.unit
    def test_evaluate_surrogate_funcs(self, alm_surr3):
        x = [0.2, 0.4, 0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]

        inputs = np.array([np.tile(x, len(x)), np.repeat(x, len(x))])
        inputs = pd.DataFrame(inputs.transpose(), columns=["x1", "x2"])

        out = alm_surr3.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out["z1"][i], rel=1e-8) == (
                2 * sin(inputs["x1"][i] ** 2)
                - 3 * cos(inputs["x2"][i] ** 3)
                - 4 * log(inputs["x1"][i] ** 4)
                + 5 * exp(inputs["x2"][i] ** 5)
            )

        # Recall evaluate_surrogate to make sure cached fcn works
        out2 = alm_surr3.evaluate_surrogate(inputs)
        for i in range(inputs.shape[0]):
            assert pytest.approx(out2["z1"][i], rel=1e-8) == (
                2 * sin(inputs["x1"][i] ** 2)
                - 3 * cos(inputs["x2"][i] ** 3)
                - 4 * log(inputs["x1"][i] ** 4)
                + 5 * exp(inputs["x2"][i] ** 5)
            )

    @pytest.mark.unit
    def test_populate_block_funcs(self, alm_surr3):
        blk = SurrogateBlock(concrete=True)

        blk.build_model(alm_surr3)

        assert isinstance(blk.inputs, Var)
        assert isinstance(blk.outputs, Var)
        assert isinstance(blk.alamo_constraint, Constraint)
        assert len(blk.alamo_constraint) == 1
        assert str(blk.alamo_constraint["z1"].body) == (
            "outputs[z1] - (2*sin(inputs[x1]**2) - 3*cos(inputs[x2]**3) - "
            "4*log(inputs[x1]**4) + 5*exp(inputs[x2]**5))"
        )

    @pytest.mark.unit
    def test_save(self, alm_surr1):
        stream = StringIO()
        alm_surr1.save(stream)
        assert stream.getvalue() == jstring

    @pytest.mark.unit
    def test_load(self):
        alm_surr = AlamoSurrogate.load(StringIO(jstring))

        x1 = 3
        x2 = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        model = alm_surr._surrogate_expressions["z1"].split("==")[1]
        model_value = eval(model)
        assert z1 == pytest.approx(model_value)
        assert alm_surr._input_labels == ["x1", "x2"]
        assert alm_surr._output_labels == ["z1"]
        assert alm_surr._input_bounds == {"x1": (0, 5), "x2": (0, 10)}

    @pytest.mark.unit
    def test_save_load(self, alm_surr1):
        with TempfileManager as tf:
            fname = tf.create_tempfile(suffix=".json")
            alm_surr1.save_to_file(fname, overwrite=True)

            assert os.path.isfile(fname)

            with open(fname, "r") as f:
                js = f.read()
            f.close()

            alm_load = AlamoSurrogate.load_from_file(fname)

        # Check file contents
        assert js == jstring

        # Check loaded object
        assert isinstance(alm_load, AlamoSurrogate)
        x1 = 3
        x2 = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        model = alm_load._surrogate_expressions["z1"].split("==")[1]
        model_value = eval(model)
        assert z1 == pytest.approx(model_value)
        assert alm_load._input_labels == ["x1", "x2"]
        assert alm_load._output_labels == ["z1"]
        assert alm_load._input_bounds == {"x1": (0, 5), "x2": (0, 10)}

        # Check for clean up
        assert not os.path.isfile(fname)

    @pytest.mark.unit
    def test_save_no_overwrite(self, alm_surr1):
        with TempfileManager as tf:
            fname = tf.create_tempfile(suffix=".json")

            with pytest.raises(FileExistsError):
                alm_surr1.save_to_file(fname)

        # Check for clean up
        assert not os.path.isfile(fname)


@pytest.mark.skipif(alamo.executable is None, reason="ALAMO not available")
@pytest.mark.integration
class TestWorkflow:
    training_data = pd.DataFrame(
        np.array(
            [
                [0.353837234435, 0.99275270941666, 0.762878272854],
                [0.904978848612, -0.746908518721, 0.387963718723],
                [0.643706630938, -0.617496599522, -0.0205375902284],
                [1.29881420688, 0.305594881575, 2.43011137696],
                [1.35791650867, 0.351045058258, 2.36989368612],
                [0.938369314089, -0.525167416293, 0.829756159423],
                [-1.46593541641, 0.383902178482, 1.14054797964],
                [-0.374378293218, -0.689730440659, -0.219122783909],
                [0.690326213554, 0.569364994374, 0.982068847698],
                [-0.961163301329, 0.499471920546, 0.936855365038],
            ]
        ),
        columns=["x1", "x2", "z1"],
    )

    @pytest.fixture(scope="class")
    def alamo_trainer(self):
        # Test end-to-end workflow with a simple problem.
        bnds = {"x1": (-1.5, 1.5), "x2": (-1.5, 1.5)}
        alamo_trainer = AlamoTrainer(
            input_labels=["x1", "x2"],
            output_labels=["z1"],
            input_bounds=bnds,
            training_dataframe=TestWorkflow.training_data,
        )

        alamo_trainer.config.linfcns = True
        alamo_trainer.config.monomialpower = [2, 3, 4, 5, 6]
        alamo_trainer.config.multi2power = [1, 2]

        (
            alamo_trainer.status,
            alamo_trainer.alamo_object,
            alamo_trainer.msg,
        ) = alamo_trainer.train_surrogate()

        return alamo_trainer

    def test_execution(self, alamo_trainer):
        # Check execution
        assert alamo_trainer.status is True
        assert alamo_trainer.msg == " Normal termination"

        # Check temp file clean up
        assert alamo_trainer._temp_context is None
        assert not os.path.exists(alamo_trainer._almfile)
        assert not os.path.exists(alamo_trainer._trcfile)

    def test_alamo_results(self, alamo_trainer):
        assert alamo_trainer._results is not None
        assert alamo_trainer._results["NINPUTS"] == "2"
        assert alamo_trainer._results["NOUTPUTS"] == "1"
        assert float(alamo_trainer._results["SSEOLR"]["z1"]) == pytest.approx(
            0.373e-29, abs=1e-10
        )
        assert float(alamo_trainer._results["SSEOLR"]["z1"]) == pytest.approx(
            0.373e-29, abs=1e-10
        )
        assert float(alamo_trainer._results["SSE"]["z1"]) == pytest.approx(
            0.976e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["RMSE"]["z1"]) == pytest.approx(
            0.988e-12, abs=1e-10
        )
        assert float(alamo_trainer._results["R2"]["z1"]) == pytest.approx(
            1.00, rel=1e-9
        )
        assert float(alamo_trainer._results["ModelSize"]["z1"]) == pytest.approx(
            6, rel=1e-9
        )
        assert float(alamo_trainer._results["BIC"]["z1"]) == pytest.approx(
            -539.0, rel=1e-9
        )
        assert float(alamo_trainer._results["RIC"]["z1"]) == pytest.approx(
            32.5, rel=1e-9
        )
        assert float(alamo_trainer._results["Cp"]["z1"]) == pytest.approx(
            2.00, rel=1e-9
        )
        assert float(alamo_trainer._results["AICc"]["z1"]) == pytest.approx(
            -513.0, rel=1e-9
        )
        assert float(alamo_trainer._results["HQC"]["z1"]) == pytest.approx(
            -543.0, rel=1e-9
        )
        assert float(alamo_trainer._results["MSE"]["z1"]) == pytest.approx(
            0.325e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["SSEp"]["z1"]) == pytest.approx(
            0.976e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["MADp"]["z1"]) == pytest.approx(
            0.115e-07, abs=1e-10
        )
        x1 = 3
        x2 = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        model = alamo_trainer._results["Model"]["z1"].split("==")[1]
        model_value = eval(model)
        assert z1 == pytest.approx(model_value)

    def test_alamo_object(self, alamo_trainer):
        alamo_object = alamo_trainer.alamo_object
        assert isinstance(alamo_object, AlamoSurrogate)
        x1 = 3
        x2 = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        model = alamo_object._surrogate_expressions["z1"].split("==")[1]
        model_value = eval(model)
        assert z1 == pytest.approx(model_value)
        assert alamo_object._input_labels == ["x1", "x2"]
        assert alamo_object._output_labels == ["z1"]
        assert alamo_object._input_bounds == {"x1": (-1.5, 1.5), "x2": (-1.5, 1.5)}

        # Check populating a block to finish workflow
        blk = SurrogateBlock(concrete=True)

        blk.build_model(alamo_object)

        assert isinstance(blk.inputs, Var)
        assert blk.inputs["x1"].bounds == (-1.5, 1.5)
        assert blk.inputs["x2"].bounds == (-1.5, 1.5)
        assert isinstance(blk.outputs, Var)
        assert blk.outputs["z1"].bounds == (None, None)
        assert isinstance(blk.alamo_constraint, Constraint)
        assert len(blk.alamo_constraint) == 1

    def test_metrics(self, alamo_trainer):
        alamo_object = alamo_trainer.alamo_object

        metrics = compute_fit_metrics(alamo_object, TestWorkflow.training_data)

        assert isinstance(metrics, dict)

        # ALAMO metrics are consistently 10x larger than they should be
        assert metrics["z1"]["SSE"] == pytest.approx(
            float(alamo_trainer._results["SSE"]["z1"]) * 0.1, rel=1e-8
        )
        assert metrics["z1"]["R2"] == pytest.approx(
            float(alamo_trainer._results["R2"]["z1"]), rel=1e-8
        )
        assert metrics["z1"]["MSE"] == pytest.approx(
            float(alamo_trainer._results["RMSE"]["z1"]) ** 2 * 0.1, rel=1e-8
        )
        assert metrics["z1"]["RMSE"] == pytest.approx(
            float(alamo_trainer._results["RMSE"]["z1"]) * 0.1, rel=1e-8
        )


@pytest.mark.skipif(alamo.executable is None, reason="ALAMO not available")
@pytest.mark.integration
class TestWorkflowValidation:
    training_data = pd.DataFrame(
        np.array(
            [
                [0.353837234435, 0.99275270941666, 0.762878272854],
                [0.904978848612, -0.746908518721, 0.387963718723],
                [0.643706630938, -0.617496599522, -0.0205375902284],
                [1.29881420688, 0.305594881575, 2.43011137696],
                [1.35791650867, 0.351045058258, 2.36989368612],
                [0.938369314089, -0.525167416293, 0.829756159423],
                [-1.46593541641, 0.383902178482, 1.14054797964],
                [-0.374378293218, -0.689730440659, -0.219122783909],
                [0.690326213554, 0.569364994374, 0.982068847698],
                [-0.961163301329, 0.499471920546, 0.936855365038],
            ]
        ),
        columns=["x1", "x2", "z1"],
    )

    @pytest.fixture(scope="class")
    def alamo_trainer(self):
        # Test end-to-end workflow with a simple problem.
        # Here we will send the training dataset as validation as well,
        # to confirm that the mechanics work
        bnds = {"x1": (-1.5, 1.5), "x2": (-1.5, 1.5)}
        alamo_trainer = AlamoTrainer(
            input_labels=["x1", "x2"],
            output_labels=["z1"],
            input_bounds=bnds,
            training_dataframe=TestWorkflow.training_data,
            validation_dataframe=TestWorkflow.training_data,
        )

        alamo_trainer.config.linfcns = True
        alamo_trainer.config.monomialpower = [2, 3, 4, 5, 6]
        alamo_trainer.config.multi2power = [1, 2]

        (
            alamo_trainer.status,
            alamo_trainer.alamo_object,
            alamo_trainer.msg,
        ) = alamo_trainer.train_surrogate()

        return alamo_trainer

    def test_execution(self, alamo_trainer):
        # Check execution
        assert alamo_trainer.status is True
        assert alamo_trainer.msg == " Normal termination"

        # Check temp file clean up
        assert alamo_trainer._temp_context is None
        assert not os.path.exists(alamo_trainer._almfile)
        assert not os.path.exists(alamo_trainer._trcfile)

    def test_alamo_results(self, alamo_trainer):
        assert alamo_trainer._results is not None
        assert alamo_trainer._results["NINPUTS"] == "2"
        assert alamo_trainer._results["NOUTPUTS"] == "1"
        assert float(alamo_trainer._results["SSEOLR"]["z1"]) == pytest.approx(
            0.373e-29, abs=1e-10
        )
        assert float(alamo_trainer._results["SSE"]["z1"]) == pytest.approx(
            0.976e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["RMSE"]["z1"]) == pytest.approx(
            0.988e-12, abs=1e-10
        )
        assert float(alamo_trainer._results["R2"]["z1"]) == pytest.approx(
            1.00, rel=1e-9
        )
        assert float(alamo_trainer._results["ModelSize"]["z1"]) == pytest.approx(
            6, rel=1e-9
        )
        assert float(alamo_trainer._results["BIC"]["z1"]) == pytest.approx(
            -539.0, rel=1e-9
        )
        assert float(alamo_trainer._results["RIC"]["z1"]) == pytest.approx(
            32.5, rel=1e-9
        )
        assert float(alamo_trainer._results["Cp"]["z1"]) == pytest.approx(
            2.00, rel=1e-9
        )
        assert float(alamo_trainer._results["AICc"]["z1"]) == pytest.approx(
            -513.0, rel=1e-9
        )
        assert float(alamo_trainer._results["HQC"]["z1"]) == pytest.approx(
            -543.0, rel=1e-9
        )
        assert float(alamo_trainer._results["MSE"]["z1"]) == pytest.approx(
            0.325e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["SSEp"]["z1"]) == pytest.approx(
            0.976e-23, abs=1e-10
        )
        assert float(alamo_trainer._results["MADp"]["z1"]) == pytest.approx(
            0.115e-07, abs=1e-10
        )
        x1 = 3
        x2 = 2
        z1 = (
            4 * x1**2
            - 4 * x2**2
            - 2.1 * x1**4
            + 4 * x2**4
            + 1 / 3.0 * x1**6
            + 1 * x1 * x2
        )
        model = alamo_trainer._results["Model"]["z1"].split("==")[1]
        model_value = eval(model)
        assert z1 == pytest.approx(model_value)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for smooth VLE formulation

Authors: Andrew Lee
"""

import pytest

from pyomo.environ import (
    ConcreteModel,
    Constraint,
    Param,
    value,
    Var,
    units as pyunits,
)

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity
from idaes.core.util.exceptions import PropertyPackageError


# Dummy EoS to use for fugacity calls
class DummyEoS(object):
    def common(self, pobj):
        pass

    def build_parameters(b):
        pass

    def fug_phase_comp(b, p, j, pp):
        return b.fug_phase_comp[p, j]


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3},
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    # Create a dummy state block
    m.props = m.params.state_block_class([1], parameters=m.params)

    m.props[1].temperature_bubble = Var([("Liq", "Vap")], initialize=300)
    m.props[1].temperature_dew = Var([("Liq", "Vap")], initialize=300)
    m.props[1]._teq = Var([("Liq", "Vap")], initialize=300)

    m.props[1].fug_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=10
    )

    SmoothVLE.phase_equil(m.props[1], ("Liq", "Vap"))

    return m


@pytest.mark.unit
def test_build(frame):
    assert isinstance(frame.props[1].eps_1_Liq_Vap, Param)
    assert value(frame.props[1].eps_1_Liq_Vap) == 0.01

    assert isinstance(frame.props[1].eps_2_Liq_Vap, Param)
    assert value(frame.props[1].eps_2_Liq_Vap) == 0.0005

    assert isinstance(frame.props[1]._t1_Liq_Vap, Var)

    assert isinstance(frame.props[1]._t1_constraint_Liq_Vap, Constraint)
    assert isinstance(frame.props[1]._teq_constraint_Liq_Vap, Constraint)


@pytest.mark.unit
def test_t1(frame):
    # Test that T1 is the max(T, T_bubble)
    # Can't check directly, but see that residual of constraint is correct
    for t in [200, 300, 400, 500]:
        for tb in [200, 300, 400, 500]:
            frame.props[1].temperature.value = t
            frame.props[1].temperature_bubble[("Liq", "Vap")].value = tb
            frame.props[1]._t1_Liq_Vap.value = max(t, tb)
            assert value(frame.props[1]._t1_constraint_Liq_Vap.body) == pytest.approx(
                0, abs=5e-3
            )


@pytest.mark.unit
def test_t_eq(frame):
    # Test that Teq is the min(T1, T_dew)
    # Can't check directly, but see that residual of constraint is correct
    for t1 in [200, 300, 400, 500]:
        for td in [200, 300, 400, 500]:
            frame.props[1]._t1_Liq_Vap.value = t1
            frame.props[1].temperature_dew[("Liq", "Vap")].value = td
            frame.props[1]._teq[("Liq", "Vap")].value = min(t1, td)
            assert value(frame.props[1]._teq_constraint_Liq_Vap.body) == pytest.approx(
                0, abs=5e-3
            )


@pytest.mark.unit
def test_non_VLE_pair():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3},
                "phase_equilibrium_form": {("Sol", "Liq"): fugacity},
            }
        },
        phases={
            "Sol": {"equation_of_state": DummyEoS},
            "Liq": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    # Create a dummy state block
    m.props = m.params.state_block_class([1], parameters=m.params)

    with pytest.raises(
        PropertyPackageError,
        match="Phase pair Liq-Sol was identified as being a VLE pair, however "
        "Liq is liquid but Sol is not vapor.",
    ):
        SmoothVLE.phase_equil(m.props[1], ("Liq", "Sol"))
